from pbpstats.resources.boxscore.boxscore import Boxscore
from pbpstats.resources.enhanced_pbp.enhanced_pbp import EnhancedPbp
from pbpstats.resources.games.games import Games
from pbpstats.resources.pbp.pbp import Pbp
from pbpstats.resources.possessions.possessions import Possessions
from pbpstats.resources.shots.shots import Shots

__all__ = [
    'Boxscore',
    'EnhancedPbp',
    'Games',
    'Pbp',
    'Possessions',
    'Shots',
]



OVERRIDES = {
    'PLAYER_NAME': 'name',
}


class StatsNbaBoxscoreItem(object):
    def __init__(self, item):
        for key, value in item.items():
            setattr(self, OVERRIDES.get(key, key.lower()), value)

    @property
    def data(self):
        return self.__dict__



import requests

from pbpstats import (
    NBA_STRING, D_LEAGUE_STRING, WNBA_STRING, NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX, WNBA_GAME_ID_PREFIX
)
from pbpstats.data_loader.abs_data_loader import AbsDataLoader


class DataNbaWebLoader(AbsDataLoader):
    """
    base class for loading data from data.nba.com
    should not be called directly
    """
    def _load_request_data(self):
        response = requests.get(self.url)
        if response.status_code == 200:
            self.source_data = response.json()
            self._save_data_to_file()
        else:
            response.raise_for_status()

    @property
    def league(self):
        """
        First 2 in game id represent league
        00 for nba, 10 for wnba, 20 for g-league
        """
        if self.game_id[0:2] == NBA_GAME_ID_PREFIX:
            return NBA_STRING
        elif self.game_id[0:2] == G_LEAGUE_GAME_ID_PREFIX:
            return D_LEAGUE_STRING  # url uses dleague instead of gleague
        elif self.game_id[0:2] == WNBA_GAME_ID_PREFIX:
            return WNBA_STRING

    @property
    def season(self):
        """
        season url part is year in which season starts
        4th and 5th characters in game id represent season year
        ex. for 2016-17 season 4th and 5th characters would be 16 and season should return 2016
        """
        if self.game_id[3] == '9':
            return '19' + self.game_id[3] + self.game_id[4]
        else:
            return '20' + self.game_id[3] + self.game_id[4]



from collections import defaultdict

import pbpstats.data_loader as data_loader


class DataLoaderFactory(object):
    def __init__(self):
        self.loaders = defaultdict(lambda: defaultdict(lambda: []))
        self._load_data_loaders()

    def _load_data_loaders(self):
        """
        loads data loaders from data_loader package
        """
        loaders = dict([(name, cls) for name, cls in data_loader.__dict__.items() if isinstance(cls, type)])
        for _, loader_cls in loaders.items():
            self.loaders[loader_cls.resource][loader_cls.data_provider].append(loader_cls)

    def get_data_loader(self, data_provider, resource):
        """
        data_provider - stats/data
        resource - name of class from resources directory
        """
        return self.loaders[resource][data_provider]



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import StatsEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.rebound import Rebound
from pbpstats.resources.enhanced_pbp.turnover import Turnover


class StatsRebound(Rebound, StatsEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    def get_offense_team_id(self):
        """
        overrides method inherited from StatsEnhancedPbpItem
        """
        if self.is_real_rebound:
            return self.missed_shot.team_id
        if isinstance(self.previous_event, Turnover) and not self.previous_event.is_no_turnover:
            # shot clock turnover has place holder rebound after turnover
            # this correct team starts next possession
            team_ids = list(self.current_players.keys())
            return team_ids[0] if team_ids[1] == self.previous_event.team_id else team_ids[1]
        return self.previous_event.get_offense_team_id()



class StatsNbaLoaderBase(object):
    """
    base class for parsing stats.nba.com api responses
    should not be called directly
    """
    def make_list_of_dicts(self, results_set_index=0):
        """
        creates list of dicts from data
        """
        headers = self.source_data['resultSets'][results_set_index]['headers']
        rows = self.source_data['resultSets'][results_set_index]['rowSet']
        deduped_rows = self.dedupe_events_row_set(rows)
        return [dict(zip(headers, row)) for row in deduped_rows]

    @staticmethod
    def dedupe_events_row_set(events_row_set):
        """
        dedupes list of list while preserving order
        used to dedupe events rowSets pbp response because some games have duplicate events
        """
        deduped_events_row_set = []
        for sublist in events_row_set:
            if sublist not in deduped_events_row_set:
                deduped_events_row_set.append(sublist)
        return deduped_events_row_set



from pbpstats.resources.base import Base
from pbpstats.resources.enhanced_pbp.field_goal import FieldGoal
from pbpstats.resources.enhanced_pbp.free_throw import FreeThrow
from pbpstats.resources.enhanced_pbp.rebound import Rebound
from pbpstats.resources.enhanced_pbp.turnover import Turnover


class EnhancedPbp(Base):
    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        return [item.data for item in self.items]

    @property
    def fgas(self):
        return [item for item in self.items if isinstance(item, FieldGoal)]

    @property
    def fgms(self):
        return [item for item in self.items if isinstance(item, FieldGoal) and item.made]

    @property
    def ftas(self):
        return [item for item in self.items if isinstance(item, FreeThrow)]

    @property
    def rebounds(self):
        return [item for item in self.items if isinstance(item, Rebound) and item.is_real_rebound]

    @property
    def turnovers(self):
        return [item for item in self.items if isinstance(item, Turnover) and not item.is_no_turnover]



from pbpstats import DEFENSIVE_GOALTENDING_STRING


class Violation(object):
    event_type = 7

    @property
    def is_delay_of_game(self):
        return self.event_action_type == 1

    @property
    def is_goaltend_violation(self):
        return self.event_action_type == 2

    @property
    def is_lane_violation(self):
        return self.event_action_type == 3

    @property
    def is_jumpball_violation(self):
        return self.event_action_type == 4

    @property
    def is_kicked_ball_violation(self):
        return self.event_action_type == 5

    @property
    def is_double_lane_violation(self):
        return self.event_action_type == 6

    @property
    def event_stats(self):
        stats = []
        if self.is_goaltend_violation:
            stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': DEFENSIVE_GOALTENDING_STRING, 'stat_value': 1})
            team_ids = list(self.current_players.keys())
            lineups_ids = self.lineup_ids
            for stat in stats:
                opponent_team_id = team_ids[0] if stat['team_id'] == team_ids[1] else team_ids[1]
                stat['lineup_id'] = lineups_ids[stat['team_id']]
                stat['opponent_team_id'] = opponent_team_id
                stat['opponent_lineup_id'] = lineups_ids[opponent_team_id]
        return self.base_stats + stats



import abc
import os
import json

from pbpstats.overrides import IntDecoder
from pbpstats.resources.enhanced_pbp.ejection import Ejection
from pbpstats.resources.enhanced_pbp.end_of_period import EndOfPeriod
from pbpstats.resources.enhanced_pbp.field_goal import FieldGoal
from pbpstats.resources.enhanced_pbp.foul import Foul
from pbpstats.resources.enhanced_pbp.free_throw import FreeThrow
from pbpstats.resources.enhanced_pbp.jump_ball import JumpBall
from pbpstats.resources.enhanced_pbp.substitution import Substitution
from pbpstats.resources.enhanced_pbp.timeout import Timeout
from pbpstats.resources.enhanced_pbp.turnover import Turnover


class InvalidNumberOfStartersException(Exception):
    pass


class StartOfPeriod(metaclass=abc.ABCMeta):
    event_type = 12

    @abc.abstractclassmethod
    def get_period_starters(self, file_directory):
        pass

    @property
    def current_players(self):
        """
        overrides EnhancedPbpItem current_players property
        """
        return self.period_starters

    def get_team_starting_with_ball(self):
        if (self.period == 1 or self.period >= 5) and isinstance(self.next_event, JumpBall):
            # period starts with jump ball - team that wins starts with the ball
            return self.next_event.team_id
        else:
            # find team id on first shot, non technical ft or turnover
            next_event = self.next_event
            while not (isinstance(next_event, (FieldGoal, Turnover)) or (isinstance(next_event, FreeThrow) and not next_event.technical_ft)):
                next_event = next_event.next_event
            return next_event.team_id

    def get_offense_team_id(self):
        """
        overrides EnhancedPbpItem offense_team_id property
        """
        return self.team_starting_with_ball

    def _get_period_starters_from_period_events(self, file_directory):
        starters = []
        subbed_in_players = []
        player_team_map = {}  # only player1 has team id in event, this is to track team
        event = self
        while event is not None and not isinstance(event, EndOfPeriod):
            if not isinstance(event, Timeout):
                player_id = event.player1_id
                if player_id != 0:
                    if not isinstance(event, JumpBall):
                        # on jump balls team id is winning team, not guaranteed to be player1 team
                        player_team_map[player_id] = event.team_id
                    if isinstance(event, Substitution):
                        player_team_map[event.player2_id] = event.team_id
                        # player_id is player going out, player2_id is playing coming in
                        if event.player2_id not in starters and event.player2_id not in subbed_in_players:
                            subbed_in_players.append(event.player2_id)
                        if player_id not in starters and player_id not in subbed_in_players:
                            starters.append(player_id)

                    is_technical_foul = isinstance(event, Foul) and (event.is_technical or event.is_double_technical)
                    if player_id not in starters and player_id not in subbed_in_players:
                        tech_ft_at_period_start = isinstance(event, FreeThrow) and event.clock == '12:00'
                        if not (is_technical_foul or isinstance(event, Ejection) or tech_ft_at_period_start):
                            # ignore all techs because a player could get a technical foul when they aren't in the game
                            starters.append(player_id)
                    # need player2_id and player3_id for players who play full period and never appear in an event as player_id - ex assists, blocks, steals, foul drawn
                    if not isinstance(event, Substitution) and not (is_technical_foul or isinstance(event, Ejection)):
                        # ignore all techs because a player could get a technical foul when they aren't in the game
                        if hasattr(event, 'player2_id') and event.player2_id not in starters and event.player2_id not in subbed_in_players:
                            starters.append(event.player2_id)
                        if hasattr(event, 'player3_id') and event.player3_id not in starters and event.player3_id not in subbed_in_players:
                            starters.append(event.player3_id)
            event = event.next_event

        # split up starters by team
        starters_by_team = {}
        dangling_starters = []  # for players who don't appear in event as player1 - won't be in player_team_map
        for player_id in starters:
            team_id = player_team_map.get(player_id)
            if team_id is not None:
                if team_id not in starters_by_team.keys():
                    starters_by_team[team_id] = []
                starters_by_team[team_id].append(player_id)
            else:
                dangling_starters.append(player_id)
        # if there is one dangling starter we can add it to team missing a starter
        if len(dangling_starters) == 1 and len(starters) == 10:
            for _, team_starters in starters_by_team.items():
                if len(team_starters) == 4:
                    team_starters += dangling_starters

        for team_id, starters in starters_by_team.items():
            if len(starters) != 5:
                # check if game and period are in overrides file
                if file_directory is None:
                    raise InvalidNumberOfStartersException(f'GameId: {self.game_id}, Period: {self.period}, TeamId: {team_id}, Players: {starters}')

                missing_period_starters_file_path = f'{file_directory}/overrides/missing_period_starters.json'
                if not os.path.isfile(missing_period_starters_file_path):
                    raise InvalidNumberOfStartersException(f'GameId: {self.game_id}, Period: {self.period}, TeamId: {team_id}, Players: {starters}')

                with open(missing_period_starters_file_path) as f:
                    # hard code corrections for games with incorrect number of starters exceptions
                    missing_period_starters = json.loads(f.read(), cls=IntDecoder)

                if (
                    self.game_id in missing_period_starters.keys() and
                    self.period in missing_period_starters[self.game_id].keys() and
                    team_id in missing_period_starters[self.game_id][self.period].keys()
                ):
                    starters_by_team[team_id] = missing_period_starters[self.game_id][self.period][team_id]
                else:
                    raise InvalidNumberOfStartersException(f'GameId: {self.game_id}, Period: {self.period}, TeamId: {team_id}, Players: {starters}')

        return starters_by_team

    @property
    def event_stats(self):
        return self.base_stats



from collections import defaultdict

from pbpstats.resources.enhanced_pbp.enhanced_pbp_item import EnhancedPbpItem
from pbpstats.resources.enhanced_pbp.start_of_period import StartOfPeriod

KEY_ATTR_MAPPER = {
    'evt': 'event_num',
    'cl': 'clock',
    'de': 'description',
    'locX': 'locX',
    'locY': 'locY',
    'opt1': 'opt1',
    'opt2': 'opt2',
    'mtype': 'event_action_type',
    'etype': 'event_type',
    'opid': 'player3_id',
    'tid': 'team_id',
    'pid': 'player1_id',
    'hs': 'home_score',
    'vs': 'away_score',
    'epid': 'player2_id',
    'oftid': 'offense_team_id',
    'ord': 'order',
}


class DataEnhancedPbpItem(EnhancedPbpItem):
    def __init__(self, item, period, game_id):
        self.game_id = game_id
        self.period = period
        for key, value in KEY_ATTR_MAPPER.items():
            attr_value = item.get(key)
            if attr_value is not None and attr_value != '':
                try:
                    # convert string values that should be ints
                    setattr(self, value, int(attr_value))
                except ValueError:
                    setattr(self, value, attr_value)
        self.player_game_fouls = defaultdict(int)
        self.possession_changing_override = False
        self.non_possession_changing_override = False
        self.score = defaultdict(int)

    @property
    def data(self):
        return self.__dict__

    @property
    def event_stats(self):
        return self.base_stats

    def get_offense_team_id(self):
        return self.offense_team_id

    @property
    def is_possession_ending_event(self):
        """
        data.pbp.com pbp has offense team id attribute
        when offense team id is different in next event, event is a possession ending event
        """
        if self.next_event is None:
            return True
        elif self.next_event is not None:
            oftid_changed = self.offense_team_id != self.next_event.offense_team_id
            # offense_team_id is 0 for first event of first quarter and overtimes - don't count it
            # sometimes there is an event before period start event that breaks things so check that too
            oftid_is_0 = self.offense_team_id == 0 or self.next_event.offense_team_id == 0
            start_of_period = isinstance(self, StartOfPeriod) or isinstance(self.next_event, StartOfPeriod)
            if oftid_changed and not oftid_is_0 and not start_of_period:
                return True
        return False



KEY_ATTR_MAPPER = {
    'GAME_ID': 'game_id',
    'GAME_DATE_EST': 'date',
    'HOME_TEAM_ID': 'home_team_id',
    'VISITOR_TEAM_ID': 'visitor_team_id',
    'GAME_STATUS_TEXT': 'status'
}


class StatsNbaGameItem(object):
    def __init__(self, item):
        for key, value in KEY_ATTR_MAPPER.items():
            if item.get(key) is not None:
                setattr(self, value, item.get(key))

    @property
    def data(self):
        return self.__dict__

    @property
    def is_final(self):
        return self.status == 'Final'



from pbpstats.data_loader.data_nba.enhanced_pbp_loader import DataNbaEnhancedPbpLoader
from pbpstats.data_loader.nba_possession_loader import NbaPossessionLoader
from pbpstats.resources.possessions.possession import Possession


class DataNbaPossessionLoader(NbaPossessionLoader):
    data_provider = 'data_nba'
    resource = 'Possessions'
    parent_object = 'Game'

    def __init__(self, game_id, source, file_directory=None):
        pbp_events = DataNbaEnhancedPbpLoader(game_id, source, file_directory)
        self.events = pbp_events.items
        events_by_possession = self._split_events_by_possession()
        self.items = [Possession(possession_events) for possession_events in events_by_possession]
        self._add_extra_attrs_to_all_possessions()



from pbpstats.resources.base import Base


class Shots(Base):
    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        return [item.data for item in self.items]



import math

import pbpstats
from pbpstats.resources.enhanced_pbp.foul import Foul
from pbpstats.resources.enhanced_pbp.free_throw import FreeThrow
from pbpstats.resources.enhanced_pbp.turnover import Turnover
from pbpstats.resources.enhanced_pbp.violation import Violation


class FieldGoal(object):
    event_type = [1, 2]

    @property
    def made(self):
        return self.event_type == 1

    @property
    def blocked(self):
        return not self.made and hasattr(self, 'player3_id')

    @property
    def assisted(self):
        return self.made and hasattr(self, 'player2_id')

    @property
    def rebound(self):
        if not self.made and self.next_event.is_real_rebound:
            return self.next_event
        return None

    @property
    def is_heave(self):
        return self.distance > pbpstats.HEAVE_DISTANCE_CUTOFF and self.seconds_remaining < pbpstats.HEAVE_TIME_CUTOFF

    @property
    def is_corner_3(self):
        if self.shot_value != 3:
            return False
        if not hasattr(self, 'locY') or self.locY is None:
            return False
        if self.locY <= 87:
            return True
        return False

    @property
    def distance(self):
        if hasattr(self, 'locX') and hasattr(self, 'locY'):
            x_squared = self.locX ** 2
            y_squared = self.locY ** 2
            # unit for distance is off by factor of 10, divide by 10 to convert to feet
            shot_distance = math.sqrt(x_squared + y_squared) / 10
            return round(shot_distance, 1)
        else:
            # no coordinates - get shot distance from event description
            try:
                return int(self.description.split("'")[0].split(' ')[-1])
            except:
                return None

    @property
    def shot_type(self):
        if self.shot_value == 3:
            if self.is_corner_3:
                return pbpstats.CORNER_3_STRING
            else:
                return pbpstats.ARC_3_STRING

        if self.distance is None:
            return pbpstats.UNKNOWN_SHOT_DISTANCE_STRING
        elif self.distance < pbpstats.AT_RIM_CUTOFF:
            return pbpstats.AT_RIM_STRING
        elif self.distance < pbpstats.SHORT_MID_RANGE_CUTOFF:
            return pbpstats.SHORT_MID_RANGE_STRING
        else:
            return pbpstats.LONG_MID_RANGE_STRING

    @property
    def putback(self):
        if self.assisted or self.shot_value == 3:
            return False
        prev_evt = self.previous_event
        if prev_evt is None:
            return False
        prev_evt_is_shooting_foul = isinstance(prev_evt, Foul) and (prev_evt.is_shooting_foul or prev_evt.is_shooting_block_foul)
        prev_evt_is_goaltend = isinstance(prev_evt, Violation) and prev_evt.is_goaltend_violation
        if (prev_evt_is_shooting_foul or prev_evt_is_goaltend) and self.clock == prev_evt.clock:
            # sometimes foul event is between rebound and shot on an and 1 or goaltend is between rebound and made shot event
            prev_evt = prev_evt.previous_event
            if prev_evt is None:
                return False
        if not hasattr(prev_evt, 'is_real_rebound'):
            return False
        if not prev_evt.is_real_rebound:
            return False
        return (
            prev_evt.oreb and
            prev_evt.player1_id == self.player1_id and
            prev_evt.seconds_remaining - self.seconds_remaining <= 2
        )

    @property
    def shot_data(self):
        team_ids = list(self.current_players.keys())
        opponent_team_id = team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
        shot_data = {
            'PlayerId': self.player1_id,
            'TeamId': self.team_id,
            'OpponentTeamId': opponent_team_id,
            'LineupId': self.lineup_ids[self.team_id],
            'OpponentLineupId': self.lineup_ids[opponent_team_id],
            'Made': self.made,
            'X': self.locX,
            'Y': self.locY,
            'Time': self.seconds_remaining,
            'ShotValue': self.shot_value,
            'Putback': self.putback,
            'ShotType': self.shot_type,
            'ScoreMargin': self.score_margin,
            'EventNum': self.event_num
        }
        if self.made:
            shot_data['Assisted'] = self.assisted
            if self.assisted:
                shot_data['AssistPlayerId'] = self.player2_id
        if not self.made:
            shot_data['Blocked'] = self.blocked
            if self.blocked:
                shot_data['BlockPlayerId'] = self.player3_id
        if self.is_second_chance_event():
            prev_event = self.previous_event
            while not (hasattr(prev_event, 'is_real_rebound') and prev_event.is_real_rebound):
                prev_event = prev_event.previous_event
            shot_data['SecondsSinceOReb'] = prev_event.seconds_remaining - self.seconds_remaining
            shot_data['OrebShotPlayerId'] = prev_event.missed_shot.player1_id
            shot_data['OrebReboundPlayerId'] = prev_event.player1_id
            if prev_event.player1_id != 0:
                rebound_shot_type = prev_event.missed_shot.shot_type
                if prev_event.missed_shot.blocked:
                    rebound_shot_type += pbpstats.BLOCKED_STRING
            elif prev_event.missed_shot.blocked:
                # separate blocked from non blocked because blocked won't have shot clock reset
                rebound_shot_type = 'TeamBlocked'
            else:
                rebound_shot_type = 'Team'
            shot_data['OrebShotType'] = rebound_shot_type
        return shot_data

    @property
    def and1(self):
        """
        note that this will return true for plays where there is a foul away from the play at the time of the shot.
        this is to make sure possession does not end if there are still FTs to shoot
        to get true and1s need to check the shooter is the player who drew the foul
        """
        # check for foul at time of shot
        shooter_team_id = self.team_id
        fouls_at_time_of_shot = []
        # ignore technical fouls and delay of game fouls when getting last foul
        events_at_shot_time = self.get_all_events_at_current_time()
        for event in events_at_shot_time:
            if isinstance(event, Foul) and not event.is_delay_of_game and not event.is_technical:
                fouls_at_time_of_shot.append(event)

        if len(fouls_at_time_of_shot) == 1:
            foul_event = fouls_at_time_of_shot[0]
            foul_team_team_id = foul_event.team_id

            if foul_event.is_flagrant:
                # flagrant foul and 1
                if foul_team_team_id != shooter_team_id:
                    return True

            other_makes_at_time_of_shot = []
            for event in events_at_shot_time:
                if isinstance(event, FieldGoal) and event.event_num != self.event_num and event.made and self.team_id == event.team_id:
                    other_makes_at_time_of_shot.append(event)
            if shooter_team_id != foul_team_team_id:
                # check FT 1 of 1s at time of shot
                ft_1_of_1s_at_time_of_shot = []
                for event in events_at_shot_time:
                    if isinstance(event, FreeThrow) and (event.ft_1_of_1 or event.ft_1pt) and not event.technical_ft:
                        ft_1_of_1s_at_time_of_shot.append(event)

                if len(other_makes_at_time_of_shot) == 1 and len(ft_1_of_1s_at_time_of_shot) == 1:
                    if ft_1_of_1s_at_time_of_shot[0].player1_id == other_makes_at_time_of_shot[0].player1_id:
                        return False
                    elif ft_1_of_1s_at_time_of_shot[0].player1_id == self.player1_id:
                        return True
                elif len(ft_1_of_1s_at_time_of_shot) > 0:
                    for ft_event in ft_1_of_1s_at_time_of_shot:
                        if ft_event.team_id == shooter_team_id:
                            return True
                else:
                    # no free throws - check for lane violation and offensive goaltending
                    for event in events_at_shot_time:
                        if isinstance(event, Turnover) and (event.is_lane_violation or event.is_offensive_goaltending):
                            return True
                        if isinstance(event, Violation) and event.is_double_lane_violation:
                            return True
        elif shooter_team_id not in [event.team_id for event in fouls_at_time_of_shot]:
            ft_1_of_1s_at_time_of_shot = []
            for event in events_at_shot_time:
                if isinstance(event, FreeThrow) and (event.ft_1_of_1 or event.ft_1pt) and not event.technical_ft:
                    ft_1_of_1s_at_time_of_shot.append(event)

            if len(ft_1_of_1s_at_time_of_shot) == 1:
                if ft_1_of_1s_at_time_of_shot[0].team_id == shooter_team_id:
                    return True
            elif len(ft_1_of_1s_at_time_of_shot) > 1:
                for ft_event in ft_1_of_1s_at_time_of_shot:
                    if ft_event.player1_id == self.player1_id:
                        return True
        else:
            opponent_fouls = [event for event in fouls_at_time_of_shot if event.team_id != shooter_team_id]
            if 1 in [event.number_of_fta_for_foul for event in opponent_fouls]:
                # rarely gets here but it is when there is an and1 and then another foul at the same time
                # for example a loose ball foul going for the rebound on the FT
                return True

        return False

    @property
    def event_stats(self):
        stats = []
        is_penalty_event = self.is_penalty_event()
        is_second_chance_event = self.is_second_chance_event()
        if self.distance is not None:
            if self.shot_value == 2:
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.TOTAL_2PT_SHOT_DISTANCE_STRING, 'stat_value': self.distance})
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.TOTAL_2PT_SHOTS_WITH_DISTANCE, 'stat_value': 1})
            else:
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.TOTAL_3PT_SHOT_DISTANCE_STRING, 'stat_value': self.distance})
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.TOTAL_3PT_SHOTS_WITH_DISTANCE, 'stat_value': 1})
                if self.is_heave:
                    if self.made:
                        stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.HEAVE_MAKES_STRING, 'stat_value': 1})
                    else:
                        stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.HEAVE_MISSES_STRING, 'stat_value': 1})

        team_ids = list(self.current_players.keys())
        opponent_team_id = team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
        if self.made and not self.assisted:
            stat_key = f'{pbpstats.UNASSISTED_STRING}{self.shot_type}'
            stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': stat_key, 'stat_value': 1})
            if is_second_chance_event:
                second_chance_stat_key = f'{pbpstats.SECOND_CHANCE_STRING}{stat_key}'
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': second_chance_stat_key, 'stat_value': 1})
            if is_penalty_event:
                penalty_stat_key = f'{pbpstats.PENALTY_STRING}{stat_key}'
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': penalty_stat_key, 'stat_value': 1})
            if self.putback:
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.PUTBACKS_STRING, 'stat_value': 1})
        elif self.assisted:
            scorer_key = f'{pbpstats.ASSISTED_STRING}{self.shot_type}'
            assist_key = f'{self.shot_type}{pbpstats.ASSISTS_STRING}'
            stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': scorer_key, 'stat_value': 1})
            stats.append({'player_id': self.player2_id, 'team_id': self.team_id, 'stat_key': assist_key, 'stat_value': 1})
            if is_second_chance_event:
                second_chance_scorer_key = f'{pbpstats.SECOND_CHANCE_STRING}{scorer_key}'
                second_chance_assist_key = f'{pbpstats.SECOND_CHANCE_STRING}{assist_key}'
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': second_chance_scorer_key, 'stat_value': 1})
                stats.append({'player_id': self.player2_id, 'team_id': self.team_id, 'stat_key': second_chance_assist_key, 'stat_value': 1})
            if is_penalty_event:
                penalty_scorer_key = f'{pbpstats.PENALTY_STRING}{scorer_key}'
                penalty_assist_key = f'{pbpstats.PENALTY_STRING}{assist_key}'
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': penalty_scorer_key, 'stat_value': 1})
                stats.append({'player_id': self.player2_id, 'team_id': self.team_id, 'stat_key': penalty_assist_key, 'stat_value': 1})
            assist_to_key = f'{self.player2_id}:AssistsTo:{self.player1_id}:{self.shot_type}'
            stats.append({'player_id': self.player2_id, 'team_id': self.team_id, 'stat_key': assist_to_key, 'stat_value': 1})
        elif self.blocked:
            shot_key = f'{self.shot_type}{pbpstats.BLOCKED_STRING}'
            stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': shot_key, 'stat_value': 1})
            block_key = f'{pbpstats.BLOCKED_STRING}{self.shot_type}'
            stats.append({'player_id': self.player3_id, 'team_id': opponent_team_id, 'stat_key': block_key, 'stat_value': 1})
            if is_second_chance_event:
                second_chance_shot_key = f'{pbpstats.SECOND_CHANCE_STRING}{shot_key}'
                second_chance_block_key = f'{pbpstats.SECOND_CHANCE_STRING}{block_key}'
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': second_chance_shot_key, 'stat_value': 1})
                stats.append({'player_id': self.player3_id, 'team_id': opponent_team_id, 'stat_key': second_chance_block_key, 'stat_value': 1})
            if is_penalty_event:
                penalty_shot_key = f'{pbpstats.PENALTY_STRING}{shot_key}'
                penalty_block_key = f'{pbpstats.PENALTY_STRING}{block_key}'
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': penalty_shot_key, 'stat_value': 1})
                stats.append({'player_id': self.player3_id, 'team_id': opponent_team_id, 'stat_key': penalty_block_key, 'stat_value': 1})
        else:
            stat_key = f'{pbpstats.MISSED_STRING}{self.shot_type}'
            stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': stat_key, 'stat_value': 1})
            if is_second_chance_event:
                second_chance_stat_key = f'{pbpstats.SECOND_CHANCE_STRING}{stat_key}'
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': second_chance_stat_key, 'stat_value': 1})
            if is_penalty_event:
                penalty_stat_key = f'{pbpstats.PENALTY_STRING}{stat_key}'
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': penalty_stat_key, 'stat_value': 1})

        if self.made:
            # add plus minus and opponent points - used for lineup/wowy stats to get net rating
            for team_id, players in self.current_players.items():
                multiplier = 1 if team_id == self.team_id else -1
                for player_id in players:
                    stat_item = {
                        'player_id': player_id,
                        'team_id': team_id,
                        'stat_key': pbpstats.PLUS_MINUS_STRING,
                        'stat_value': self.shot_value * multiplier,
                    }
                    stats.append(stat_item)
                    if multiplier == -1:
                        opponent_points_stat_item = {
                            'player_id': player_id,
                            'team_id': team_id,
                            'stat_key': pbpstats.OPPONENT_POINTS,
                            'stat_value': self.shot_value,
                        }
                        stats.append(opponent_points_stat_item)

        lineups_ids = self.lineup_ids
        for stat in stats:
            opponent_team_id = team_ids[0] if stat['team_id'] == team_ids[1] else team_ids[1]
            stat['lineup_id'] = lineups_ids[stat['team_id']]
            stat['opponent_team_id'] = opponent_team_id
            stat['opponent_lineup_id'] = lineups_ids[opponent_team_id]

        return self.base_stats + stats



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import DataEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.free_throw import FreeThrow


class DataFreeThrow(FreeThrow, DataEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    @property
    def made(self):
        return ' Missed' not in self.description



from pbpstats.resources.base import Base


class Games(Base):
    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        return [item.data for item in self.items]

    @property
    def final_games(self):
        return [item.data for item in self.items if item.is_final]



import json
import os

from pbpstats import NBA_STRING, D_LEAGUE_STRING
from pbpstats.data_loader.data_nba.file_loader import DataNbaFileLoader
from pbpstats.data_loader.data_nba.web_loader import DataNbaWebLoader
from pbpstats.resources.pbp.data_nba_pbp_item import DataNbaPbpItem


class DataNbaPbpLoader(DataNbaFileLoader, DataNbaWebLoader):
    data_provider = 'data_nba'
    resource = 'Pbp'
    parent_object = 'Game'

    def __init__(self, game_id, source, file_directory=None):
        self.game_id = game_id
        self.file_directory = file_directory
        self.source = source
        self._load_data()
        self._make_pbp_items()

    def _load_data(self):
        source_method = getattr(self, f'_from_{self.source}')
        source_method()

    def _from_file(self):
        self.file_path = f'{self.file_directory}/pbp/data_{self.game_id}.json'
        self._load_data_from_file()

    def _from_web(self):
        league_url_part = NBA_STRING if self.league == D_LEAGUE_STRING else self.league
        self.url = f'https://data.{league_url_part}.com/data/v2015/json/mobile_teams/{self.league}/{self.season}/scores/pbp/{self.game_id}_full_pbp.json'
        self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f'{self.file_directory}/pbp/data_{self.game_id}.json'
            with open(file_path, 'w') as outfile:
                json.dump(self.source_data, outfile)

    def _make_pbp_items(self):
        self.items = [DataNbaPbpItem(event, item['p']) for item in self.data for event in item['pla']]

    @property
    def data(self):
        return self.source_data['g']['pd']



from collections import defaultdict

import requests

from pbpstats import HEADERS, REQUEST_TIMEOUT
from pbpstats.resources.enhanced_pbp.enhanced_pbp_item import EnhancedPbpItem
from pbpstats.resources.enhanced_pbp.field_goal import FieldGoal
from pbpstats.resources.enhanced_pbp.foul import Foul
from pbpstats.resources.enhanced_pbp.free_throw import FreeThrow
from pbpstats.resources.enhanced_pbp.jump_ball import JumpBall
from pbpstats.resources.enhanced_pbp.rebound import Rebound
from pbpstats.resources.enhanced_pbp.start_of_period import StartOfPeriod
from pbpstats.resources.enhanced_pbp.turnover import Turnover
from pbpstats.resources.enhanced_pbp.violation import Violation


KEY_ATTR_MAPPER = {
    'GAME_ID': 'game_id',
    'EVENTNUM': 'event_num',
    'PCTIMESTRING': 'clock',
    'PERIOD': 'period',
    'EVENTMSGACTIONTYPE': 'event_action_type',
    'EVENTMSGTYPE': 'event_type',
    'PLAYER1_ID': 'player1_id',
    'PLAYER1_TEAM_ID': 'team_id',
    'PLAYER2_ID': 'player2_id',
    'PLAYER3_ID': 'player3_id',
    'VIDEO_AVAILABLE_FLAG': 'video_available',
}


class StatsEnhancedPbpItem(EnhancedPbpItem):
    def __init__(self, event, order):
        for key, value in KEY_ATTR_MAPPER.items():
            if event.get(key) is not None:
                setattr(self, value, event.get(key))

        if event.get('HOMEDESCRIPTION') is not None and event.get('VISITORDESCRIPTION') is not None:
            self.description = f"{event.get('HOMEDESCRIPTION')}: {event.get('VISITORDESCRIPTION')}"
        elif event.get('HOMEDESCRIPTION') is not None:
            self.description = f"{event.get('HOMEDESCRIPTION')}"
        elif event.get('VISITORDESCRIPTION') is not None:
            self.description = f"{event.get('VISITORDESCRIPTION')}"
        elif event.get('NEUTRALDESCRIPTION') is not None:
            self.description = f"{event.get('NEUTRALDESCRIPTION')}"
        else:
            self.description = ''

        if event.get('PLAYER1_TEAM_ID') is None and event.get('PLAYER1_ID') is not None:
            # need to set team id in these cases where player id is team id
            self.team_id = event.get('PLAYER1_ID', 0)
            self.player1_id = 0

        # fix team/player ids on some event types so they are consistent with DataPbpItem
        if self.event_type == 10:
            # jump ball PLAYER3_TEAM_ID is player who ball gets tipped to
            self.player2_id = event['PLAYER3_ID']
            self.player3_id = event['PLAYER2_ID']
            if event['PLAYER3_TEAM_ID'] is not None:
                self.team_id = event['PLAYER3_TEAM_ID']
            else:
                # when jump ball is tipped out of bounds, winning team is PLAYER3_ID
                self.team_id = event['PLAYER3_ID']
                if hasattr(self, 'player2_id'):
                    delattr(self, 'player2_id')
        elif self.event_type in [5, 6]:
            # steals need to change PLAYER2_ID to player3_id - this is player who turned ball over
            # fouls need to change PLAYER2_ID to player3_id - this is player who drew foul
            if hasattr(self, 'player2_id'):
                delattr(self, 'player2_id')
            if event.get('PLAYER2_ID') is not None:
                self.player3_id = event['PLAYER2_ID']

        if hasattr(self, 'player2_id') and self.player2_id == 0:
            delattr(self, 'player2_id')
        if hasattr(self, 'player3_id') and self.player3_id == 0:
            delattr(self, 'player3_id')
        self.order = order
        self.player_game_fouls = defaultdict(int)
        self.possession_changing_override = False
        self.non_possession_changing_override = False
        self.score = defaultdict(int)

    @property
    def data(self):
        return self.__dict__

    @property
    def video_url(self):
        """
        gets url for mp4 video of play
        """
        if self.video_available == 1:
            parameters = {
                'GameEventID': self.event_num,
                'GameID': self.game_id
            }
            base_url = 'https://stats.nba.com/stats/videoeventsasset'
            response = requests.get(base_url, parameters, headers=HEADERS, timeout=REQUEST_TIMEOUT)
            if response.status_code == 200:
                response_json = response.json()
                video_urls = response_json['resultSets']['Meta']['videoUrls']
                if len(video_urls) == 1:
                    return video_urls[0]['murl']
            else:
                response.raise_for_status()
        return None

    def get_offense_team_id(self):
        """
        this method gets overriden in FieldGoal, FreeThrow, Rebound, Turnover, JumpBall
        """
        if isinstance(self, Foul) and (self.is_charge or self.is_offensive_foul):
            # offensive foul returns team id
            # this isn't separate method in Foul class because some fouls can be committed
            # on offense or defense (loose ball, flagrant, technical)
            return self.team_id
        event_to_check = self.previous_event
        team_ids = list(self.current_players.keys())
        while (
            event_to_check is not None and
            not (
                isinstance(event_to_check, (FieldGoal, JumpBall)) or
                (isinstance(event_to_check, Turnover) and not event_to_check.is_no_turnover) or
                (isinstance(event_to_check, Rebound) and event_to_check.is_real_rebound) or
                (isinstance(event_to_check, FreeThrow) and not event_to_check.technical_ft)
            )
        ):
            event_to_check = event_to_check.previous_event
        if event_to_check is None and self.next_event is not None:
            # should only get here on first possession of period when first event is non-offensive foul,
            # FieldGoal, FreeThrow, Rebound, Turnover, JumpBall
            return self.next_event.get_offense_team_id()
        if isinstance(event_to_check, Turnover) and not event_to_check.is_no_turnover:
            return team_ids[0] if team_ids[1] == event_to_check.get_offense_team_id() else team_ids[1]
        if isinstance(event_to_check, Rebound) and event_to_check.is_real_rebound:
            if not event_to_check.oreb:
                return team_ids[0] if team_ids[1] == event_to_check.get_offense_team_id() else team_ids[1]
            return event_to_check.get_offense_team_id()
        if isinstance(event_to_check, (FieldGoal, FreeThrow)):
            if event_to_check.is_possession_ending_event:
                return team_ids[0] if team_ids[1] == event_to_check.get_offense_team_id() else team_ids[1]
            return event_to_check.get_offense_team_id()
        if isinstance(event_to_check, JumpBall):
            if event_to_check.count_as_possession:
                team_ids = list(self.current_players.keys())
                return team_ids[0] if team_ids[1] == event_to_check.get_offense_team_id() else team_ids[1]
            return event_to_check.get_offense_team_id()

    @property
    def is_possession_ending_event(self):
        if self.next_event is None:
            return True

        if self.possession_changing_override:
            return True

        if self.non_possession_changing_override:
            return False

        if isinstance(self, Rebound) and self.is_real_rebound and not self.oreb:
            return True

        if isinstance(self, Turnover) and not self.is_no_turnover:
            return True

        if isinstance(self, FieldGoal) and self.made:
            # no possession change on flagrant foul
            next_event_is_flagrant_drawn = (
                isinstance(self.next_event, Foul) and
                self.next_event.is_flagrant and
                self.team_id != self.next_event.team_id and
                self.clock == self.next_event.clock
            )
            if not self.and1 and not next_event_is_flagrant_drawn:
                return True

        if isinstance(self, FreeThrow) and self.made and self.end_ft:
            next_event_is_foul_drawn_at_ft_time = (
                isinstance(self.next_event, Foul) and
                self.clock == self.next_event.clock and
                self.team_id != self.next_event.team_id and
                (
                    self.next_event.is_loose_ball_foul or
                    self.next_event.is_personal_foul or
                    self.next_event.is_away_from_play_foul or
                    self.next_event.is_flagrant
                )
            )
            if not self.is_away_from_play_ft and not self.is_inbound_foul_ft and not next_event_is_foul_drawn_at_ft_time:
                return True

        if not isinstance(self.previous_event, StartOfPeriod) and isinstance(self, JumpBall):
            # need to check for rare case where possession changes on jump ball but there is no turnover/rebound
            if isinstance(self.next_event, Turnover) and not self.next_event.is_no_turnover and self.next_event.clock == self.clock:
                # if next event is steal at same time of pbp don't need to change possession since steal takes care of it
                return False
            elif isinstance(self.previous_event, Turnover) and not self.previous_event.is_no_turnover and self.previous_event.clock == self.clock:
                # if previous event is steal at same time of pbp don't need to change possession since steal takes care of it
                return False
            elif isinstance(self.next_event, Violation) and self.next_event.is_jumpball_violation:
                # jump ball violation - turnover will be possession changing event
                return False
            elif isinstance(self.next_event, Foul) and self.next_event.clock == self.clock:
                next_event = self.next_event.next_event
                if isinstance(next_event, Turnover) and not next_event.is_no_turnover and next_event.clock == self.clock:
                    # foul turnover on jump ball - turnover will trigger change of possession
                    return False

            jump_ball_winning_team_id = self.team_id

            prev_event = self.previous_event
            while prev_event is not None and not prev_event.is_possession_ending_event:
                prev_event = prev_event.previous_event

            if prev_event is None:
                return False

            if isinstance(prev_event, Rebound):
                jump_ball_winning_team_started_possession_with_ball = jump_ball_winning_team_id == prev_event.team_id
            else:
                jump_ball_winning_team_started_possession_with_ball = jump_ball_winning_team_id != prev_event.team_id

            next_event_rebound = isinstance(self.next_event, Rebound) and self.next_event.is_real_rebound

            if not jump_ball_winning_team_started_possession_with_ball and not(next_event_rebound or isinstance(self.next_event, JumpBall)):
                # ignore jump ball if next event is a rebound or jump ball since that will trigger possession change
                return True

        return False



import abc

import pbpstats
from pbpstats.resources.enhanced_pbp.field_goal import FieldGoal
from pbpstats.resources.enhanced_pbp.foul import Foul
from pbpstats.resources.enhanced_pbp.free_throw import FreeThrow
from pbpstats.resources.enhanced_pbp.rebound import Rebound


class EnhancedPbpItem(metaclass=abc.ABCMeta):
    def __repr__(self):
        return f"<{type(self).__name__} GameId: {self.game_id}, Description: {self.description}, Time: {self.clock}, EventNum: {self.event_num}>"

    @abc.abstractproperty
    def is_possession_ending_event(self):
        pass

    @abc.abstractproperty
    def event_stats(self):
        pass

    @abc.abstractmethod
    def get_offense_team_id(self):
        pass

    @property
    def base_stats(self):
        """
        these are event stats for all event types
        - seconds played off/def
        - possession off/def
        """
        return self._get_seconds_played_stats_items() + self._get_possessions_played_stats_items()

    def get_all_events_at_current_time(self):
        events = [self]
        # going backwards
        event = self
        while event is not None and self.seconds_remaining == event.seconds_remaining:
            if event != self:
                events.append(event)
            event = event.previous_event
        # going forwards
        event = self
        while event is not None and self.seconds_remaining == event.seconds_remaining:
            if event != self:
                events.append(event)
            event = event.next_event
        return sorted(events, key=lambda k: k.order)

    @property
    def seconds_remaining(self):
        split = self.clock.split(':')
        return float(split[0]) * 60 + float(split[1])

    @property
    def current_players(self):
        """
        for all non subsitution events current players are just
        the same as previous event
        this gets overwritten in Substitution class
        """
        return self.previous_event.current_players

    @property
    def score_margin(self):
        """
        from perspective of offense team
        """
        if self.previous_event is None:
            score = self.score
        else:
            score = self.previous_event.score
        offense_team_id = self.get_offense_team_id()
        offense_points = score[offense_team_id]
        defense_points = 0
        for team_id, points in score.items():
            if team_id != offense_team_id:
                defense_points = points
        return offense_points - defense_points

    @property
    def lineup_ids(self):
        """
        hyphen separated sorted player id strings
        """
        lineup_ids = {}
        for team_id, team_players in self.current_players.items():
            players = [str(player_id) for player_id in team_players]
            sorted_player_ids = sorted(players)
            lineup_id = '-'.join(sorted_player_ids)
            lineup_ids[team_id] = lineup_id
        return lineup_ids

    @property
    def seconds_since_previous_event(self):
        if self.previous_event is None:
            return 0
        return self.previous_event.seconds_remaining - self.seconds_remaining

    def is_second_chance_event(self):
        event = self.previous_event
        if isinstance(event, Rebound) and event.is_real_rebound and event.oreb:
            return True
        while not (event is None or event.is_possession_ending_event):
            if isinstance(event, Rebound) and event.is_real_rebound and event.oreb:
                return True
            event = event.previous_event
        return False

    def is_penalty_event(self):
        if hasattr(self, 'fouls_to_give'):
            team_ids = list(self.current_players.keys())
            offense_team_id = self.get_offense_team_id()
            defense_team_id = team_ids[0] if offense_team_id == team_ids[1] else team_ids[1]
            if self.fouls_to_give[defense_team_id] == 0:
                if isinstance(self, (Foul, FreeThrow, Rebound)):
                    # if foul or free throw or rebound on a missed ft
                    # check foul event and should return false is foul
                    # was shooting foul and team had a foul to give
                    if isinstance(self, Foul):
                        foul_event = self
                    elif isinstance(self, FreeThrow):
                        foul_event = self.foul_that_led_to_ft
                    else:
                        # if rebound is on missed ft, also need to look at foul that led to FT
                        if not self.oreb and isinstance(self.missed_shot, FreeThrow):
                            foul_event = self.missed_shot.foul_that_led_to_ft
                        else:
                            return True
                    if foul_event is None:
                        return True
                    fouls_to_give_prior_to_foul = foul_event.previous_event.fouls_to_give[defense_team_id]
                    if fouls_to_give_prior_to_foul > 0:
                        return False
                return True
        return False

    @property
    def count_as_possession(self):
        """
        don't count possession change if it starts with <= 2 seconds left
        and no points are scored before period ends
        """
        if self.is_possession_ending_event:
            if self.seconds_remaining > 2:
                return True
            # check when previous possession ended
            prev_event = self.previous_event
            while prev_event is not None and not prev_event.is_possession_ending_event:
                prev_event = prev_event.previous_event
            if prev_event is None or prev_event.seconds_remaining > 2:
                return True
            # possession starts in final 2 seconds
            # return True if there is a FT or FGM between now and end of period
            next_event = prev_event.next_event
            while next_event is not None:
                if isinstance(next_event, FreeThrow) or (isinstance(next_event, FieldGoal) and next_event.made):
                    return True
                next_event = next_event.next_event
        return False

    def _get_seconds_played_stats_items(self):
        """
        makes event stats items for:
        - seconds played
        - seconds played for number of fouls
        - second chance seconds played
        - penalty seconds played
        """
        stat_items = []
        team_ids = list(self.current_players.keys())
        offense_team_id = self.get_offense_team_id()
        is_penalty_event = self.is_penalty_event()
        is_second_chance_event = self.is_second_chance_event()
        if self.seconds_since_previous_event != 0:
            for team_id, players in self.previous_event.current_players.items():
                seconds_stat_key = (
                    pbpstats.SECONDS_PLAYED_OFFENSE_STRING
                    if team_id == offense_team_id
                    else pbpstats.SECONDS_PLAYED_DEFENSE_STRING
                )
                opponent_team_id = team_ids[0] if team_id == team_ids[1] else team_ids[1]
                previous_poss_lineup_ids = self.previous_event.lineup_ids
                for player_id in players:
                    keys_to_add = [seconds_stat_key]
                    player_fouls = self.previous_event.player_game_fouls[player_id]
                    period = self.period if self.period <= 4 else 'OT'
                    foul_tracking_seconds_stat_key = f'Period{period}Fouls{player_fouls}{seconds_stat_key}'
                    keys_to_add.append(foul_tracking_seconds_stat_key)
                    if is_second_chance_event:
                        seconds_chance_seconds_stat_key = f'{pbpstats.SECOND_CHANCE_STRING}{seconds_stat_key}'
                        keys_to_add.append(seconds_chance_seconds_stat_key)
                    if is_penalty_event:
                        penalty_seconds_stat_key = f'{pbpstats.PENALTY_STRING}{seconds_stat_key}'
                        keys_to_add.append(penalty_seconds_stat_key)
                    for stat_key in keys_to_add:
                        stat_item = {
                            'player_id': player_id,
                            'team_id': team_id,
                            'opponent_team_id': opponent_team_id,
                            'lineup_id': previous_poss_lineup_ids[team_id],
                            'opponent_lineup_id': previous_poss_lineup_ids[opponent_team_id],
                            'stat_key': stat_key,
                            'stat_value': self.seconds_since_previous_event,
                        }
                        stat_items.append(stat_item)
        return stat_items

    def _get_possessions_played_stats_items(self):
        """
        makes event stats items for:
        - possessions played
        - second chance possessions played
        - penalty possessions played
        """
        stat_items = []
        team_ids = list(self.current_players.keys())
        offense_team_id = self.get_offense_team_id()
        is_penalty_event = self.is_penalty_event()
        is_second_chance_event = self.is_second_chance_event()
        if self.count_as_possession:
            if isinstance(self, FreeThrow):
                current_players = self.event_for_efficiency_stats.current_players
                lineup_ids = self.event_for_efficiency_stats.lineup_ids
            else:
                current_players = self.current_players
                lineup_ids = self.lineup_ids
            for team_id, players in current_players.items():
                possessions_stat_key = (
                    pbpstats.OFFENSIVE_POSSESSION_STRING
                    if team_id == offense_team_id
                    else pbpstats.DEFENSIVE_POSSESSION_STRING
                )
                opponent_team_id = team_ids[0] if team_id == team_ids[1] else team_ids[1]
                for player_id in players:
                    keys_to_add = [possessions_stat_key]
                    if is_second_chance_event:
                        seconds_chance_possessions_stat_key = f'{pbpstats.SECOND_CHANCE_STRING}{possessions_stat_key}'
                        keys_to_add.append(seconds_chance_possessions_stat_key)
                    if is_penalty_event:
                        penalty_possessions_stat_key = f'{pbpstats.PENALTY_STRING}{possessions_stat_key}'
                        keys_to_add.append(penalty_possessions_stat_key)
                    for stat_key in keys_to_add:
                        stat_item = {
                            'player_id': player_id,
                            'team_id': team_id,
                            'opponent_team_id': opponent_team_id,
                            'lineup_id': lineup_ids[team_id],
                            'opponent_lineup_id': lineup_ids[opponent_team_id],
                            'stat_key': stat_key,
                            'stat_value': 1,
                        }
                        stat_items.append(stat_item)

        return stat_items



KEY_ATTR_MAPPER = {
    'gid': 'game_id',
    'gdte': 'date',
    'stt': 'status',
}


class DataNbaGameItem(object):
    def __init__(self, item):
        for key, value in KEY_ATTR_MAPPER.items():
            if item.get(key) is not None:
                setattr(self, value, item.get(key))
        self.home_team_id = item['h']['tid']
        self.home_team_abbreviation = item['h']['ta']
        self.home_score = item['h']['s']
        self.away_team_id = item['v']['tid']
        self.away_team_abbreviation = item['v']['ta']
        self.away_score = item['v']['s']

    @property
    def data(self):
        return self.__dict__

    @property
    def is_final(self):
        return self.status == 'Final'



from itertools import groupby
from operator import itemgetter

import pbpstats
from pbpstats.resources.enhanced_pbp.field_goal import FieldGoal
from pbpstats.resources.enhanced_pbp.free_throw import FreeThrow
from pbpstats.resources.enhanced_pbp.jump_ball import JumpBall
from pbpstats.resources.enhanced_pbp.rebound import Rebound
from pbpstats.resources.enhanced_pbp.substitution import Substitution
from pbpstats.resources.enhanced_pbp.timeout import Timeout
from pbpstats.resources.enhanced_pbp.turnover import Turnover


class Possession(object):
    def __init__(self, events):
        self.game_id = events[0].game_id
        self.period = events[0].period
        self.events = events

    def __repr__(self):
        return (
            f'<{type(self).__name__} GameId: {self.game_id}, Period: {self.period}, '
            f'Number: {self.number}, StartTime: {self.start_time}, EndTime: {self.end_time}, '
            f'OffenseTeamId: {self.offense_team_id}>'
        )

    @property
    def data(self):
        return self.__dict__

    @property
    def start_time(self):
        if not hasattr(self, 'previous_possession') or self.previous_possession is None:
            return self.events[0].clock
        return self.previous_possession.events[-1].clock

    @property
    def end_time(self):
        return self.events[-1].clock

    @property
    def start_score_margin(self):
        if not hasattr(self, 'previous_possession') or self.previous_possession is None:
            score = self.events[0].score
        else:
            score = self.previous_possession.events[-1].score
        offense_team_id = self.offense_team_id
        offense_points = score[offense_team_id]
        defense_points = 0
        for team_id, points in score.items():
            if team_id != offense_team_id:
                defense_points = points
        return offense_points - defense_points

    @property
    def offense_team_id(self):
        if len(self.events) == 1 and isinstance(self.events[0], JumpBall):
            # if possession only has one event and it is a jump ball, need to check
            # how previous possession ended to see which team actually started with the ball
            # because team id on jump ball is team that won the jump ball
            prev_event = self.previous_possession_ending_event
            if isinstance(prev_event, Turnover) and not prev_event.is_no_turnover:
                team_ids = [self.previous_possession.offense_team_id, self.previous_possession.previous_possession.offense_team_id]
                return team_ids[0] if team_ids[1] == prev_event.get_offense_team_id() else team_ids[1]
            if isinstance(prev_event, Rebound) and prev_event.is_real_rebound:
                if not prev_event.oreb:
                    team_ids = list(set([event.team_id for event in self.events] + [event.team_id for event in self.previous_possession.events]))
                    return team_ids[0] if team_ids[1] == prev_event.get_offense_team_id() else team_ids[1]
                return prev_event.get_offense_team_id()
            if isinstance(prev_event, (FieldGoal, FreeThrow)):
                if prev_event.made:
                    team_ids = list(set([event.team_id for event in self.events] + [event.team_id for event in self.previous_possession.events]))
                    return team_ids[0] if team_ids[1] == prev_event.get_offense_team_id() else team_ids[1]
                return prev_event.get_offense_team_id()
        return self.events[0].get_offense_team_id()

    @property
    def possession_has_timeout(self):
        """
        checks if there was a timeout called on the current possession
        """
        for i, event in enumerate(self.events):
            if isinstance(event, Timeout) and event.clock != self.end_time:
                # timeout is not at possession end time
                if not (
                    event.next_event is not None and
                    (isinstance(event.next_event, FreeThrow) and not event.next_event.technical_ft) and
                    event.clock == event.next_event.clock
                ):
                    # check to make sure timeout is not between/before FTs
                    return True
            elif isinstance(event, Timeout) and event.clock == self.end_time:
                timeout_time = event.clock
                after_timeout_index = i + 1
                # call time out and turn ball over at same time as timeout following time out
                for possession_event in self.events[after_timeout_index:]:
                    if isinstance(possession_event, Turnover) and not possession_event.is_no_turnover and possession_event.clock == timeout_time:
                        return True
        return False

    @property
    def previous_possession_has_timeout(self):
        """
        check for timeout on previous possession - if there is a timeout at the same time as possession end, current possession starts off timeout
        """
        if self.previous_possession is not None:
            for event in self.previous_possession.events:
                if isinstance(event, Timeout) and event.clock == self.start_time:
                    if not (
                        event.next_event is not None and
                        isinstance(event.next_event, FreeThrow) and
                        event.clock == event.next_event.clock
                    ):
                        # check to make sure timeout is not beween FTs
                        return True
        return False

    @property
    def previous_possession_ending_event(self):
        """
        gets previous possession ending event - ignoring subs
        """
        previous_event_index = -1
        while isinstance(self.previous_possession.events[previous_event_index], Substitution) and len(self.previous_possession.events) > abs(previous_event_index):
            previous_event_index -= 1
        return self.previous_possession.events[previous_event_index]

    @property
    def possession_start_type(self):
        if self.number == 1:
            return pbpstats.OFF_DEADBALL_STRING
        if self.possession_has_timeout or self.previous_possession_has_timeout:
            return pbpstats.OFF_TIMEOUT_STRING
        previous_possession_ending_event = self.previous_possession_ending_event
        if isinstance(previous_possession_ending_event, (FieldGoal, FreeThrow)) and previous_possession_ending_event.made:
            shot_type = previous_possession_ending_event.shot_type
            return f'Off{shot_type}{pbpstats.MAKE_STRING}'
        if isinstance(previous_possession_ending_event, Turnover):
            if previous_possession_ending_event.is_steal:
                return pbpstats.OFF_LIVE_BALL_TURNOVER_STRING
            return pbpstats.OFF_DEADBALL_STRING
        if isinstance(previous_possession_ending_event, Rebound):
            if previous_possession_ending_event.player1_id == 0:
                # team rebound
                return pbpstats.OFF_DEADBALL_STRING
            missed_shot = previous_possession_ending_event.missed_shot
            shot_type = missed_shot.shot_type
            if hasattr(missed_shot, 'blocked') and missed_shot.blocked:
                return f'Off{shot_type}{pbpstats.BLOCK_STRING}'
            return f'Off{shot_type}{pbpstats.MISS_STRING}'

        if isinstance(previous_possession_ending_event, JumpBall):
            # jump balls tipped out of bounds have no player2_id and should be off deadball
            if not hasattr(previous_possession_ending_event, 'player2_id'):
                return pbpstats.OFF_LIVE_BALL_TURNOVER_STRING
            else:
                return pbpstats.OFF_DEADBALL_STRING
        return pbpstats.OFF_DEADBALL_STRING

    @property
    def previous_possession_end_shooter_player_id(self):
        if self.previous_possession is not None and not(self.possession_has_timeout or self.previous_possession_has_timeout):
            previous_possession_ending_event = self.previous_possession_ending_event
            if isinstance(previous_possession_ending_event, FieldGoal) and previous_possession_ending_event.made:
                return previous_possession_ending_event.player1_id
            if isinstance(previous_possession_ending_event, Rebound):
                if previous_possession_ending_event.player1_id != 0:
                    missed_shot = previous_possession_ending_event.missed_shot
                    return missed_shot.player1_id
        return 0

    @property
    def previous_possession_end_rebound_player_id(self):
        if self.previous_possession is not None and not(self.possession_has_timeout or self.previous_possession_has_timeout):
            previous_possession_ending_event = self.previous_possession_ending_event
            if isinstance(previous_possession_ending_event, Rebound):
                if previous_possession_ending_event.player1_id != 0:
                    return previous_possession_ending_event.player1_id
        return 0

    @property
    def previous_possession_end_turnover_player_id(self):
        if self.previous_possession is not None and not(self.possession_has_timeout or self.previous_possession_has_timeout):
            previous_possession_ending_event = self.previous_possession_ending_event
            if isinstance(previous_possession_ending_event, Turnover):
                if previous_possession_ending_event.is_steal:
                    return previous_possession_ending_event.player1_id
        return 0

    @property
    def previous_possession_end_steal_player_id(self):
        if self.previous_possession is not None and not(self.possession_has_timeout or self.previous_possession_has_timeout):
            previous_possession_ending_event = self.previous_possession_ending_event
            if isinstance(previous_possession_ending_event, Turnover):
                if previous_possession_ending_event.is_steal:
                    return previous_possession_ending_event.player3_id
        return 0

    @property
    def possession_stats(self):
        grouper = itemgetter('player_id', 'team_id', 'opponent_team_id', 'lineup_id', 'opponent_lineup_id', 'stat_key')
        results = []
        event_stats = [event_stat for event in self.events for event_stat in event.event_stats]
        for key, group in groupby(sorted(event_stats, key=grouper), grouper):
            temp_dict = dict(zip(['player_id', 'team_id', 'opponent_team_id', 'lineup_id', 'opponent_lineup_id', 'stat_key'], key))
            temp_dict['stat_value'] = sum(item['stat_value'] for item in group)
            results.append(temp_dict)

        return results



class Substitution(object):
    event_type = 8

    @property
    def outgoing_player_id(self):
        return self.player1_id

    @property
    def incoming_player_id(self):
        return self.player2_id

    @property
    def current_players(self):
        """
        overrides EnhancedPbpItem current_players property
        """
        players = self.previous_event.current_players.copy()
        players[self.team_id] = [self.incoming_player_id if player == self.outgoing_player_id else player for player in players[self.team_id]]
        return players

    @property
    def event_stats(self):
        return self.base_stats



from pbpstats.resources.enhanced_pbp.stats_nba.ejection import StatsEjection
from pbpstats.resources.enhanced_pbp.stats_nba.end_of_period import StatsEndOfPeriod
from pbpstats.resources.enhanced_pbp.stats_nba.field_goal import StatsFieldGoal
from pbpstats.resources.enhanced_pbp.stats_nba.foul import StatsFoul
from pbpstats.resources.enhanced_pbp.stats_nba.free_throw import StatsFreeThrow
from pbpstats.resources.enhanced_pbp.stats_nba.jump_ball import StatsJumpBall
from pbpstats.resources.enhanced_pbp.stats_nba.rebound import StatsRebound
from pbpstats.resources.enhanced_pbp.stats_nba.replay import StatsReplay
from pbpstats.resources.enhanced_pbp.stats_nba.start_of_period import StatsStartOfPeriod
from pbpstats.resources.enhanced_pbp.stats_nba.substitution import StatsSubstitution
from pbpstats.resources.enhanced_pbp.stats_nba.timeout import StatsTimeout
from pbpstats.resources.enhanced_pbp.stats_nba.turnover import StatsTurnover
from pbpstats.resources.enhanced_pbp.stats_nba.violation import StatsViolation

__all__ = [
    'StatsEjection',
    'StatsEndOfPeriod',
    'StatsFieldGoal',
    'StatsFoul',
    'StatsFreeThrow',
    'StatsJumpBall',
    'StatsRebound',
    'StatsReplay',
    'StatsStartOfPeriod',
    'StatsSubstitution',
    'StatsTimeout',
    'StatsTurnover',
    'StatsViolation',
]



from itertools import groupby
from operator import itemgetter

from pbpstats import KEYS_OFF_BY_FACTOR_OF_5_WHEN_AGGREGATING_FOR_TEAM_AND_LINEUPS
from pbpstats.resources.base import Base


class Possessions(Base):
    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        return self.__dict__

    def _aggregate_event_stats(self, *args):
        stats = [event_stat for item in self.items for event in item.events for event_stat in event.event_stats]
        grouper = itemgetter(*args)
        results = []
        for key, grp in groupby(sorted(stats, key=grouper), grouper):
            temp_dict = dict(zip([*args], key))
            value = sum(item['stat_value'] for item in grp)
            if temp_dict['stat_key'] in KEYS_OFF_BY_FACTOR_OF_5_WHEN_AGGREGATING_FOR_TEAM_AND_LINEUPS and 'player_id' not in args:
                # since stat keys are summed up from player stats
                # team and lineup stats will need some stats to be divided by 5
                value = value / 5
            temp_dict['stat_value'] = value if isinstance(value, int) else round(value, 1)
            results.append(temp_dict)
        return results

    @property
    def team_stats(self):
        return self._aggregate_event_stats('team_id', 'stat_key')

    @property
    def opponent_stats(self):
        return self._aggregate_event_stats('opponent_team_id', 'stat_key')

    @property
    def player_stats(self):
        return self._aggregate_event_stats('player_id', 'team_id', 'stat_key')

    @property
    def lineup_stats(self):
        return self._aggregate_event_stats('lineup_id', 'team_id', 'stat_key')

    @property
    def lineup_opponent_stats(self):
        return self._aggregate_event_stats('opponent_lineup_id', 'opponent_team_id', 'stat_key')



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import StatsEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.substitution import Substitution


class StatsSubstitution(Substitution, StatsEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)



class StatsNbaPbpItem(object):
    def __init__(self, event, order):
        for key, value in event.items():
            setattr(self, key.lower(), value)
        self.order = order

    @property
    def data(self):
        return self.__dict__



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import DataEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.replay import Replay


class DataReplay(Replay, DataEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import StatsEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.foul import Foul


class StatsFoul(Foul, StatsEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    @property
    def number_of_fta_for_foul(self):
        clock = self.clock
        event = self
        while event is not None and event.clock == clock and not (hasattr(event, 'first_ft') and not event.technical_ft and self.team_id != event.team_id):
            event = event.next_event

        if event is not None and hasattr(event, 'first_ft') and not event.technical_ft and event.clock == clock and (not hasattr(self, 'player3_id') or self.player3_id == event.player1_id):
            # player3 id check is to make sure player who got fouled is player shooting free throws, prior to 2005-06 because foul drawning player isn't in pbp
            if 'of 1' in event.description:
                return 1
            elif 'of 2' in event.description:
                return 2
            elif 'of 3' in event.description:
                return 3

        # if we haven't found ft yet, try going backwards
        event = self
        while event is not None and event.clock == clock and not (hasattr(event, 'first_ft') and not event.technical_ft and self.team_id != event.team_id):
            event = event.previous_event

        if event is not None and hasattr(event, 'first_ft') and not event.technical_ft and event.clock == clock and (not hasattr(self, 'player3_id') or self.player3_id == event.player1_id):
            # player3 id check is to make sure player who got fouled is player shooting free throws, prior to 2005-06 because foul drawning player isn't in pbp
            if 'of 1' in event.description:
                return 1
            elif 'of 2' in event.description:
                return 2
            elif 'of 3' in event.description:
                return 3
        return None



from pbpstats.data_loader.stats_nba.pbp_loader import StatsNbaPbpLoader
from pbpstats.data_loader.stats_nba.shots_loader import StatsNbaShotsLoader
from pbpstats.data_loader.nba_enhanced_pbp_loader import NbaEnhancedPbpLoader
from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_factory import StatsNbaEnhancedPbpFactory
from pbpstats.resources.enhanced_pbp.field_goal import FieldGoal
from pbpstats.resources.enhanced_pbp.rebound import EventOrderError


class StatsNbaEnhancedPbpLoader(StatsNbaPbpLoader, NbaEnhancedPbpLoader):
    data_provider = 'stats_nba'
    resource = 'EnhancedPbp'
    parent_object = 'Game'

    def __init__(self, game_id, source, file_directory=None):
        super().__init__(game_id, source, file_directory)

    def _make_pbp_items(self):
        self.factory = StatsNbaEnhancedPbpFactory()
        self.items = [self.factory.get_event_class(item['EVENTMSGTYPE'])(item, i) for i, item in enumerate(self.data)]
        self._add_extra_attrs_to_all_events()
        self._check_rebound_event_order(6)
        self._add_shot_x_y_coords()

    def _add_shot_x_y_coords(self):
        shots_loader = StatsNbaShotsLoader(self.game_id, self.source, self.file_directory)
        shots_event_num_map = {item.game_event_id: {'loc_x': item.loc_x, 'loc_y': item.loc_y} for item in shots_loader.items}
        for item in self.items:
            if isinstance(item, FieldGoal) and item.event_num in shots_event_num_map.keys():
                item.locX = shots_event_num_map[item.event_num]['loc_x']
                item.locY = shots_event_num_map[item.event_num]['loc_y']

    def _check_rebound_event_order(self, max_retries):
        """
        checks rebound events to make sure they are ordered correctly
        """
        attempts = 0
        while attempts <= max_retries:
            try:
                for event in self.items:
                    if hasattr(event, 'missed_shot'):
                        event.missed_shot
                break
            except EventOrderError as e:
                self._fix_common_event_order_error(e)
                self.items = [self.factory.get_event_class(item['EVENTMSGTYPE'])(item, i) for i, item in enumerate(self.data)]
                self._add_extra_attrs_to_all_events()
                attempts += 1

    def _fix_common_event_order_error(self, exception):
        """
        fixs common cases where events are out of order
        current cases are:
        - subs/timeouts between free throws being between missed second FT and rebound
        - end of period replay events being between missed shot and rebound
        - rebound and shot in reverse order
        - shot, rebound, rebound - first rebound needs to me moved to before shot
        - shot, rebound, rebound - second rebound needs to be moved ahead of shot and first rebound
        """
        event_num = int(str(exception).split('EventNum: ')[-1].split('>')[0])
        headers = self.source_data['resultSets'][0]['headers']
        rows = self.source_data['resultSets'][0]['rowSet']
        event_num_index = headers.index('EVENTNUM')
        event_type_index = headers.index('EVENTMSGTYPE')
        for i, row in enumerate(rows):
            if row[event_num_index] == int(event_num):
                issue_event_index = i

        if rows[issue_event_index][event_type_index] in [8, 9]:
            # these are subs/timeouts that are in between missed ft and rebound
            # move all sub/timeout events between ft and rebound to before ft
            row_index = issue_event_index
            while rows[row_index][event_type_index] in [8, 9]:
                row_index -= 1

            # rows[row_index] should be moved to right before rows[issue_event_index]
            new_rows = []
            row_to_move_event_num = rows[row_index][event_num_index]
            for row in rows:
                if row[event_num_index] == row_to_move_event_num:
                    row_to_move = row
                elif row[event_num_index] == int(event_num):
                    new_rows.append(row)
                    new_rows.append(row_to_move)
                else:
                    new_rows.append(row)

            self.source_data['resultSets'][0]['rowSet'] = new_rows
        elif rows[issue_event_index][event_type_index] == 18 and rows[issue_event_index + 1][event_type_index] == 4:
            # these are replays that are in between missed shot and rebound
            # move replay event after rebound
            replay_event = rows[issue_event_index]
            rebound_event = rows[issue_event_index + 1]

            self.source_data['resultSets'][0]['rowSet'][issue_event_index + 1] = replay_event
            self.source_data['resultSets'][0]['rowSet'][issue_event_index] = rebound_event

        elif rows[issue_event_index + 1][event_type_index] == 4 and rows[issue_event_index + 1][event_num_index] == int(event_num) - 1:
            # rebound and shot need to be flipped
            rebound_event = rows[issue_event_index + 1]
            shot_event = rows[issue_event_index]
            self.source_data['resultSets'][0]['rowSet'][issue_event_index] = rebound_event
            self.source_data['resultSets'][0]['rowSet'][issue_event_index + 1] = shot_event
        elif (
                rows[issue_event_index + 1][event_type_index] == 4 and rows[issue_event_index + 1][event_num_index] == int(event_num) + 2 and
                rows[issue_event_index - 1][event_type_index] == 2 and rows[issue_event_index - 1][event_num_index] == int(event_num) + 1
        ):
            # shot, rebound, rebound - first rebound need to me moved to before shot
            rebound_event = rows[issue_event_index]
            shot_event = rows[issue_event_index - 1]
            self.source_data['resultSets'][0]['rowSet'][issue_event_index] = shot_event
            self.source_data['resultSets'][0]['rowSet'][issue_event_index - 1] = rebound_event
        elif (
                rows[issue_event_index + 1][event_type_index] == 4 and rows[issue_event_index + 1][event_num_index] == int(event_num) - 2 and
                rows[issue_event_index - 1][event_type_index] == 2 and rows[issue_event_index - 1][event_num_index] == int(event_num) - 1
        ):
            # shot, rebound, rebound - second rebound needs to be moved ahead of shot and first rebound
            first_rebound = rows[issue_event_index + 1]
            second_rebound = rows[issue_event_index]
            shot_event = rows[issue_event_index - 1]
            self.source_data['resultSets'][0]['rowSet'][issue_event_index + 1] = second_rebound
            self.source_data['resultSets'][0]['rowSet'][issue_event_index] = shot_event
            self.source_data['resultSets'][0]['rowSet'][issue_event_index - 1] = first_rebound

        self._save_data_to_file()



from pbpstats.resources.enhanced_pbp.data_nba.ejection import DataEjection
from pbpstats.resources.enhanced_pbp.data_nba.end_of_period import DataEndOfPeriod
from pbpstats.resources.enhanced_pbp.data_nba.field_goal import DataFieldGoal
from pbpstats.resources.enhanced_pbp.data_nba.foul import DataFoul
from pbpstats.resources.enhanced_pbp.data_nba.free_throw import DataFreeThrow
from pbpstats.resources.enhanced_pbp.data_nba.jump_ball import DataJumpBall
from pbpstats.resources.enhanced_pbp.data_nba.rebound import DataRebound
from pbpstats.resources.enhanced_pbp.data_nba.replay import DataReplay
from pbpstats.resources.enhanced_pbp.data_nba.start_of_period import DataStartOfPeriod
from pbpstats.resources.enhanced_pbp.data_nba.substitution import DataSubstitution
from pbpstats.resources.enhanced_pbp.data_nba.timeout import DataTimeout
from pbpstats.resources.enhanced_pbp.data_nba.turnover import DataTurnover
from pbpstats.resources.enhanced_pbp.data_nba.violation import DataViolation

__all__ = [
    'DataEjection',
    'DataEndOfPeriod',
    'DataFieldGoal',
    'DataFoul',
    'DataFreeThrow',
    'DataJumpBall',
    'DataRebound',
    'DataReplay',
    'DataStartOfPeriod',
    'DataSubstitution',
    'DataTimeout',
    'DataTurnover',
    'DataViolation',
]



import pbpstats
from pbpstats.resources.enhanced_pbp.substitution import Substitution


class Turnover(object):
    event_type = 5

    @property
    def is_steal(self):
        return hasattr(self, 'player3_id')

    @property
    def is_no_turnover(self):
        return self.event_action_type == 0

    @property
    def is_bad_pass(self):
        return self.event_action_type == 1 and self.is_steal

    @property
    def is_lost_ball(self):
        return self.event_action_type == 2 and self.is_steal

    @property
    def is_travel(self):
        return self.event_action_type == 4

    @property
    def is_3_second_violation(self):
        return self.event_action_type == 8

    @property
    def is_shot_clock_violation(self):
        return self.event_action_type == 11

    @property
    def is_offensive_goaltending(self):
        return self.event_action_type == 15

    @property
    def is_lane_violation(self):
        return self.event_action_type == 17

    @property
    def is_kicked_ball(self):
        return self.event_action_type == 19

    @property
    def is_step_out_of_bounds(self):
        return self.event_action_type == 39

    @property
    def is_lost_ball_out_of_bounds(self):
        # some labelled as lost ball but should be lost ball out of bounds (missing player3 id)
        return self.event_action_type == 40 or (self.event_action_type == 2 and not self.is_steal)

    @property
    def is_bad_pass_out_of_bounds(self):
        # some labelled as bad pass but should be bad pass out of bounds (missing player3 id)
        return self.event_action_type == 45 or (self.event_action_type == 1 and not self.is_steal)

    @property
    def event_stats(self):
        stats = []
        second_chance_stats = []
        if not self.is_no_turnover:
            team_ids = list(self.current_players.keys())
            opponent_team_id = team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
            lineup_ids = self.lineup_ids
            if self.is_steal:
                turnover_key = pbpstats.LOST_BALL_TURNOVER_STRING if self.is_lost_ball else pbpstats.BAD_PASS_TURNOVER_STRING
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': turnover_key, 'stat_value': 1})
                steal_key = pbpstats.LOST_BALL_STEAL_STRING if self.is_lost_ball else pbpstats.BAD_PASS_STEAL_STRING
                stats.append({'player_id': self.player3_id, 'team_id': opponent_team_id, 'stat_key': steal_key, 'stat_value': 1})
            else:
                stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.DEADBALL_TURNOVERS_STRING, 'stat_value': 1})
                if self.is_travel:
                    stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.TRAVELS_STRING, 'stat_value': 1})
                elif self.is_3_second_violation:
                    stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.THREE_SECOND_VIOLATION_TURNOVER_STRING, 'stat_value': 1})
                elif self.is_step_out_of_bounds:
                    stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.STEP_OUT_OF_BOUNDS_TURNOVER_STRING, 'stat_value': 1})
                elif self.is_offensive_goaltending:
                    stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.OFFENSIVE_GOALTENDING_STRING, 'stat_value': 1})
                elif self.is_lost_ball_out_of_bounds:
                    stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.LOST_BALL_OUT_OF_BOUNDS_TURNOVER_STRING, 'stat_value': 1})
                elif self.is_bad_pass_out_of_bounds:
                    stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.BAD_PASS_OUT_OF_BOUNDS_TURNOVER_STRING, 'stat_value': 1})
                elif self.is_shot_clock_violation:
                    stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': pbpstats.SHOT_CLOCK_VIOLATION_TURNOVER_STRING, 'stat_value': 1})

                # sometimes events are out of order and turnover happens after sub
                # if player who turned ball over is subbed out before turnover event, fix lineup id
                events_to_check = self.get_all_events_at_current_time()
                if self.player1_id != 0 and str(self.player1_id) not in lineup_ids[self.team_id].split('-'):
                    # sub in wrong order - fix lineup
                    for event in events_to_check:
                        if isinstance(event, Substitution):
                            if event.outgoing_player_id == self.player1_id:
                                fixed_lineup_id = lineup_ids[self.team_id].replace(str(event.incoming_player_id), str(event.outgoing_player_id))
                                lineup_ids[self.team_id] = fixed_lineup_id

            for stat in stats:
                opponent_team_id = team_ids[0] if stat['team_id'] == team_ids[1] else team_ids[1]
                stat['lineup_id'] = lineup_ids[stat['team_id']]
                stat['opponent_team_id'] = opponent_team_id
                stat['opponent_lineup_id'] = lineup_ids[opponent_team_id]
                if self.is_second_chance_event():
                    second_chance_stats.append({key: value if key != 'stat_key' else f'{pbpstats.SECOND_CHANCE_STRING}{value}' for key, value in stat.items()})
                if self.is_penalty_event():
                    second_chance_stats.append({key: value if key != 'stat_key' else f'{pbpstats.PENALTY_STRING}{value}' for key, value in stat.items()})

        return self.base_stats + stats + second_chance_stats



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import StatsEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.free_throw import FreeThrow


class StatsFreeThrow(FreeThrow, StatsEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    @property
    def made(self):
        return 'MISS ' not in self.description

    def get_offense_team_id(self):
        """
        overrides method inherited from StatsEnhancedPbpItem
        """
        return self.team_id



class DataNbaPbpItem(object):
    def __init__(self, item, period):
        self.period = period
        for key, value in item.items():
            setattr(self, key, value)

    @property
    def data(self):
        return self.__dict__



from pbpstats.data_loader.data_nba.pbp_loader import DataNbaPbpLoader
from pbpstats.data_loader.nba_enhanced_pbp_loader import NbaEnhancedPbpLoader
from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_factory import DataNbaEnhancedPbpFactory


class DataNbaEnhancedPbpLoader(DataNbaPbpLoader, NbaEnhancedPbpLoader):
    data_provider = 'data_nba'
    resource = 'EnhancedPbp'
    parent_object = 'Game'

    def __init__(self, game_id, source, file_directory=None):
        super().__init__(game_id, source, file_directory)

    def _make_pbp_items(self):
        factory = DataNbaEnhancedPbpFactory()
        self.items = [factory.get_event_class(event['etype'])(event, item['p'], self.game_id) for item in self.data for event in item['pla']]
        self._add_extra_attrs_to_all_events()



OVERRIDES = {
    'pid': 'player_id'
}


class DataNbaBoxscoreItem(object):
    def __init__(self, item, team_id=None, team_abbreviation=None):
        self.team_id = team_id
        self.team_abbreviation = team_abbreviation
        for key, value in item.items():
            setattr(self, OVERRIDES.get(key, key), value)
        if hasattr(self, 'fn') and hasattr(self, 'ln'):
            self.name = f'{self.fn} {self.ln}'

    @property
    def data(self):
        return self.__dict__



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import DataEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.start_of_period import StartOfPeriod


class DataStartOfPeriod(StartOfPeriod, DataEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    def get_period_starters(self, file_directory=None):
        return self._get_period_starters_from_period_events(file_directory)



import re

import pbpstats.objects as objects
import pbpstats.resources as resources
from pbpstats.data_loader.factory import DataLoaderFactory


DATA_LOADER_SUFFIX = 'DataLoaderClass'
DATA_SOURCE_SUFFIX = 'DataSource'
PATTERN = re.compile(r'(?<!^)(?=[A-Z])')  # for converting camel case to snake case


class Client(object):
    def __init__(self, settings):
        data_loader = DataLoaderFactory()
        self.settings = settings
        self.data_directory = settings.get('dir')
        self._load_objects()
        self._load_resources()
        for resource, value in settings.items():
            if resource in data_loader.loaders.keys():
                resource_data_loaders = data_loader.get_data_loader(value['data_provider'], resource)
                for resource_data_loader in resource_data_loaders:
                    parent_cls_name = resource_data_loader.parent_object
                    setattr(getattr(self, parent_cls_name), f'{resource}{DATA_LOADER_SUFFIX}', resource_data_loader)
                    setattr(getattr(self, parent_cls_name), f'{resource}{DATA_SOURCE_SUFFIX}', value['source'])
                    setattr(getattr(self, parent_cls_name), resource, self.resource_dict[resource])

    def _load_objects(self):
        """
        loads classes from objects package
        """
        object_dict = dict([(name, cls) for name, cls in objects.__dict__.items() if isinstance(cls, type)])
        for name, object_cls in object_dict.items():
            setattr(self, name, object_cls)
            setattr(getattr(self, name), 'data_directory', self.data_directory)

    def _load_resources(self):
        """
        loads classes from resources package
        """
        self.resource_dict = dict([(name, cls) for name, cls in resources.__dict__.items() if isinstance(cls, type)])



import requests

from pbpstats import (
    NBA_STRING, G_LEAGUE_STRING, WNBA_STRING, NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX, WNBA_GAME_ID_PREFIX
)
from pbpstats import HEADERS, REQUEST_TIMEOUT
from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import StatsEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.start_of_period import StartOfPeriod, InvalidNumberOfStartersException


class StatsStartOfPeriod(StartOfPeriod, StatsEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    def get_period_starters(self, file_directory=None):
        try:
            return self._get_period_starters_from_period_events(file_directory)
        except InvalidNumberOfStartersException:
            return self._get_starters_from_boxscore_request()

    def _get_starters_from_boxscore_request(self):
        """
        makes request to boxscore url for time from period start to first event to get period starters
        """
        base_url = f'https://stats.{self.league_url_part}.com/stats/boxscoretraditionalv2'
        event = self
        while event is not None and event.seconds_remaining == self.seconds_remaining:
            event = event.next_event
        seconds_to_first_event = self.seconds_remaining - event.seconds_remaining

        if self.period == 1:
            start_range = 0
        elif self.period <= 4:
            start_range = int(7200 * (self.period - 1))
        else:
            start_range = int(28800 + 3000 * (self.period - 5))
        end_range = int(start_range + seconds_to_first_event * 10)
        params = {
            'GameId': self.game_id,
            'StartPeriod': 0,
            'EndPeriod': 0,
            'RangeType': 2,
            'StartRange': start_range,
            'EndRange': end_range
        }
        starters_by_team = {}
        response = requests.get(base_url, params, headers=HEADERS, timeout=REQUEST_TIMEOUT)
        if response.status_code == 200:
            response_json = response.json()
        else:
            response.raise_for_status()

        headers = response_json['resultSets'][0]['headers']
        rows = response_json['resultSets'][0]['rowSet']
        players = [dict(zip(headers, row)) for row in rows]
        starters = sorted(players, key=lambda k: int(k['MIN'].split(':')[1]), reverse=True)

        for starter in starters[0:10]:
            team_id = starter['TEAM_ID']
            player_id = starter['PLAYER_ID']
            if team_id not in starters_by_team.keys():
                starters_by_team[team_id] = []
            starters_by_team[team_id].append(player_id)

        for team_id, starters in starters_by_team.items():
            if len(starters) != 5:
                raise InvalidNumberOfStartersException(f'GameId: {self.game_id}, Period: {self.period}, TeamId: {team_id}, Players: {starters}')

        return starters_by_team

    @property
    def league_url_part(self):
        if self.game_id[0:2] == NBA_GAME_ID_PREFIX:
            return NBA_STRING
        elif self.game_id[0:2] == G_LEAGUE_GAME_ID_PREFIX:
            return f'{G_LEAGUE_STRING}.{NBA_STRING}'
        elif self.game_id[0:2] == WNBA_GAME_ID_PREFIX:
            return WNBA_STRING



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import DataEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.foul import Foul


class DataFoul(Foul, DataEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    @property
    def number_of_fta_for_foul(self):
        if '(1 FTA)' in self.description:
            return 1
        elif '(2 FTA)' in self.description:
            return 2
        elif '(3 FTA)' in self.description:
            return 3



from pbpstats.resources.base import Base


class Boxscore(Base):
    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        return {'player': self.player_items, 'team': self.team_items}

    @property
    def player_items(self):
        return [item.data for item in self.items if hasattr(item, 'player_id')]

    @property
    def team_items(self):
        return [item.data for item in self.items if not hasattr(item, 'player_id')]

    @property
    def player_name_map(self):
        return {item['player_id']: item['name'] for item in self.player_items}

    @property
    def player_team_map(self):
        return {item['player_id']: item['team_id'] for item in self.player_items}



import os
import json

from pbpstats.overrides import IntDecoder
from pbpstats.data_loader.stats_nba.enhanced_pbp_loader import StatsNbaEnhancedPbpLoader
from pbpstats.data_loader.nba_possession_loader import NbaPossessionLoader
from pbpstats.resources.possessions.possession import Possession
from pbpstats.resources.enhanced_pbp.foul import Foul


class TeamHasBackToBackPossessionsException(Exception):
    pass


class StatsNbaPossessionLoader(NbaPossessionLoader):
    data_provider = 'stats_nba'
    resource = 'Possessions'
    parent_object = 'Game'

    def __init__(self, game_id, source, file_directory=None):

        self.file_directory = file_directory
        self.game_id = game_id
        pbp_events = StatsNbaEnhancedPbpLoader(game_id, source, file_directory)
        self.events = pbp_events.items
        events_by_possession = self._split_events_by_possession()
        self.items = [Possession(possession_events) for possession_events in events_by_possession]
        self._add_extra_attrs_to_all_possessions()
        self._load_bad_possession_overrides()
        self._check_that_possessions_alternate()

    def _check_that_possessions_alternate(self):
        """
        checks that a team doesn't have back-to-back possessions
        usually caused by pbp events being out of order that can be fixed manually
        """
        for possession in self.items:
            if possession.previous_possession is not None:
                poss = possession
                prev_poss = possession.previous_possession
            elif possession.next_possession is not None:
                poss = possession.next_possession
                prev_poss = possession
            if poss.offense_team_id == prev_poss.offense_team_id:
                if not (
                    self.game_id in self.bad_pbp_cases.keys() and
                    poss.period in self.bad_pbp_cases[self.game_id].keys() and
                    poss.number in self.bad_pbp_cases[self.game_id][poss.period]
                ):
                    ignore_because_of_flagrant = False
                    events_to_check = [event for event in prev_poss.events]
                    if prev_poss.previous_possession is not None:
                        events_to_check += prev_poss.previous_possession.events
                    for event in events_to_check:
                        if isinstance(event, Foul) and event.is_flagrant:
                            ignore_because_of_flagrant = True

                    if not ignore_because_of_flagrant:
                        exception_text = (
                            f'GameId: {poss.game_id}, Period: {poss.period}, '
                            f'Number: {poss.number}, Events: {poss.events}, '
                            f'Previous Events: {prev_poss.events}>'
                        )

                        raise TeamHasBackToBackPossessionsException(exception_text)

    def _load_bad_possession_overrides(self):
        self.bad_pbp_cases = {}
        if self.file_directory is not None:
            bad_pbp_possessions_file_path = f'{self.file_directory}/overrides/bad_pbp_possessions.json'
            if os.path.isfile(bad_pbp_possessions_file_path):
                with open(bad_pbp_possessions_file_path) as f:
                    # bad pbp where event is missing in pbp causing back to back possessions for same team - this will prevent back to back possession exception from being raised
                    # {GameId: {Period:[EventNum]}}
                    self.bad_pbp_cases = json.loads(f.read(), cls=IntDecoder)



class StatsNbaShot(object):
    def __init__(self, item):
        for key, value in item.items():
            setattr(self, key.lower(), value)

    @property
    def data(self):
        return self.__dict__



import json
import os

from pbpstats import (
    NBA_STRING, G_LEAGUE_STRING, WNBA_STRING, NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX, WNBA_GAME_ID_PREFIX
)
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader
from pbpstats.resources.games.stats_nba_game_item import StatsNbaGameItem


class StatsNbaScoreboardLoader(StatsNbaFileLoader, StatsNbaWebLoader):
    data_provider = 'stats_nba'
    resource = 'Games'
    parent_object = 'Day'

    def __init__(self, date, league_string, file_directory, source):
        self.date = date
        self.league_string = league_string
        self.file_directory = file_directory
        self.source = source
        self._load_data()
        self._make_scoreboard_items()

    def _load_data(self):
        source_method = getattr(self, f'_from_{self.source}')
        source_method()

    def _from_file(self):
        self.file_path = f'{self.file_directory}/schedule/stats_{self.league_string}_{self.date.replace("/", "_")}.json'
        self._load_data_from_file()

    def _from_web(self):
        league_url_part = f'{G_LEAGUE_STRING}.{NBA_STRING}' if self.league_string == G_LEAGUE_STRING else self.league_string
        self.base_url = f'https://stats.{league_url_part}.com/stats/scoreboardV2'
        self.parameters = {
            'DayOffset': 0,
            'LeagueID': self.league_id,
            'gameDate': self.date
        }
        self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f'{self.file_directory}/schedule/stats_{self.league_string}_{self.date.replace("/", "_")}.json'
            with open(file_path, 'w') as outfile:
                json.dump(self.source_data, outfile)

    @property
    def league_id(self):
        if self.league_string == NBA_STRING:
            return NBA_GAME_ID_PREFIX
        elif self.league_string == WNBA_STRING:
            return WNBA_GAME_ID_PREFIX
        elif self.league_string == G_LEAGUE_STRING:
            return G_LEAGUE_GAME_ID_PREFIX

    def _make_scoreboard_items(self):
        self.items = [StatsNbaGameItem(item) for item in self.data]



import pbpstats


class Foul(object):
    event_type = 6

    @property
    def is_personal_foul(self):
        return self.event_action_type == 1

    @property
    def is_shooting_foul(self):
        return self.event_action_type == 2

    @property
    def is_loose_ball_foul(self):
        return self.event_action_type == 3

    @property
    def is_offensive_foul(self):
        return self.event_action_type == 4

    @property
    def is_inbound_foul(self):
        return self.event_action_type == 5

    @property
    def is_away_from_play_foul(self):
        return self.event_action_type == 6

    @property
    def is_clear_path_foul(self):
        return self.event_action_type == 9

    @property
    def is_double_foul(self):
        return self.event_action_type == 10

    @property
    def is_technical(self):
        return self.event_action_type in [11, 12, 13, 18, 19, 25, 30]

    @property
    def is_flagrant(self):
        return self.event_action_type in [14, 15]

    @property
    def is_double_technical(self):
        return self.event_action_type == 16

    @property
    def is_defensive_3_seconds(self):
        return self.event_action_type == 17

    @property
    def is_delay_of_game(self):
        return self.event_action_type == 18

    @property
    def is_charge(self):
        return self.event_action_type == 26

    @property
    def is_personal_block_foul(self):
        return self.event_action_type == 27

    @property
    def is_personal_take_foul(self):
        return self.event_action_type == 28

    @property
    def is_shooting_block_foul(self):
        return self.event_action_type == 29

    @property
    def counts_towards_penalty(self):
        return self.event_action_type in [1, 2, 3, 5, 6, 9, 14, 15, 27, 28, 29]

    @property
    def counts_as_personal_foul(self):
        return self.event_action_type in [1, 2, 3, 4, 5, 6, 9, 10, 14, 15, 26, 27, 28, 29]

    @property
    def foul_type_string(self):
        if self.is_personal_foul:
            return pbpstats.PERSONAL_FOUL_TYPE_STRING
        if self.is_shooting_foul:
            return pbpstats.SHOOTING_FOUL_TYPE_STRING
        if self.is_loose_ball_foul:
            return pbpstats.LOOSE_BALL_FOUL_TYPE_STRING
        if self.is_offensive_foul:
            return pbpstats.OFFENSIVE_FOUL_TYPE_STRING
        if self.is_inbound_foul:
            return pbpstats.INBOUND_FOUL_TYPE_STRING
        if self.is_away_from_play_foul:
            return pbpstats.AWAY_FROM_PLAY_FOUL_TYPE_STRING
        if self.is_clear_path_foul:
            return pbpstats.CLEAR_PATH_FOUL_TYPE_STRING
        if self.is_double_foul:
            return pbpstats.DOUBLE_FOUL_TYPE_STRING
        if self.event_action_type == 14:
            return pbpstats.FLAGRANT_1_FOUL_TYPE_STRING
        if self.event_action_type == 15:
            return pbpstats.FLAGRANT_2_FOUL_TYPE_STRING
        if self.is_defensive_3_seconds:
            return pbpstats.DEFENSIVE_3_SECONDS_FOUL_TYPE_STRING
        if self.is_charge:
            return pbpstats.CHARGE_FOUL_TYPE_STRING
        if self.is_personal_block_foul:
            return pbpstats.PERSONAL_BLOCK_TYPE_STRING
        if self.is_personal_take_foul:
            return pbpstats.PERSONAL_TAKE_TYPE_STRING
        if self.is_shooting_block_foul:
            return pbpstats.SHOOTING_BLOCK_TYPE_STRING

    @property
    def event_stats(self):
        stats = []
        foul_type = self.foul_type_string
        if foul_type is not None:
            stats.append({'player_id': self.player1_id, 'team_id': self.team_id, 'stat_key': foul_type, 'stat_value': 1})
            team_ids = list(self.current_players.keys())
            if hasattr(self, 'player3_id'):
                p3_stat_key = foul_type if self.is_double_foul else foul_type + pbpstats.FOULS_DRAWN_TYPE_STRING
                opponent_team_id = team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
                if self.player3_id in self.current_players[self.team_id]:
                    stats.append({'player_id': self.player3_id, 'team_id': self.team_id, 'stat_key': p3_stat_key, 'stat_value': 1})
                elif self.player3_id in self.current_players[opponent_team_id]:
                    stats.append({'player_id': self.player3_id, 'team_id': opponent_team_id, 'stat_key': p3_stat_key, 'stat_value': 1})

            lineups_ids = self.lineup_ids
            for stat in stats:
                opponent_team_id = team_ids[0] if stat['team_id'] == team_ids[1] else team_ids[1]
                stat['lineup_id'] = lineups_ids[stat['team_id']]
                stat['opponent_team_id'] = opponent_team_id
                stat['opponent_lineup_id'] = lineups_ids[opponent_team_id]
        return self.base_stats + stats



import requests

from pbpstats import (
    NBA_STRING, G_LEAGUE_STRING, WNBA_STRING, PLAYOFFS_STRING, REGULAR_SEASON_STRING,
    NBA_GAME_ID_PREFIX, G_LEAGUE_GAME_ID_PREFIX, WNBA_GAME_ID_PREFIX
)
from pbpstats import HEADERS, REQUEST_TIMEOUT
from pbpstats.data_loader.abs_data_loader import AbsDataLoader
from pbpstats.data_loader.stats_nba.base import StatsNbaLoaderBase


class StatsNbaWebLoader(AbsDataLoader, StatsNbaLoaderBase):
    """
    base class for loading data from stats.nba.com api endpoint
    should not be called directly
    """
    def _load_request_data(self):
        response = requests.get(self.base_url, self.parameters, headers=HEADERS, timeout=REQUEST_TIMEOUT)
        if response.status_code == 200:
            self.source_data = response.json()
            self._save_data_to_file()
        else:
            response.raise_for_status()

    @property
    def data(self):
        return self.make_list_of_dicts()

    @property
    def league(self):
        """
        First 2 in game id represent league
        00 for nba, 10 for wnba, 20 for g-league
        """
        if self.game_id[0:2] == NBA_GAME_ID_PREFIX:
            return NBA_STRING
        elif self.game_id[0:2] == G_LEAGUE_GAME_ID_PREFIX:
            return G_LEAGUE_STRING
        elif self.game_id[0:2] == WNBA_GAME_ID_PREFIX:
            return WNBA_STRING

    @property
    def season(self):
        """
        4th and 5th characters in game id represent season year
        ex. for 2016-17 season 4th and 5th characters would be 16
        for WNBA, season is just year, ex. 2016
        """
        digit4 = int(self.game_id[3])
        digit5 = int(self.game_id[4])
        if digit4 == 9:
            if digit5 == 9:
                return '1999' if self.league == WNBA_STRING else '1999-00'
            else:
                return f'19{digit4}{digit5}' if self.league == WNBA_STRING else f'19{digit4}{digit5}-{digit4}{digit5 + 1}'
        elif digit5 == 9:
            return f'20{digit4}{digit5}' if self.league == WNBA_STRING else f'20{digit4}{digit5}-{digit4 + 1}0'
        else:
            return f'20{digit4}{digit5}'if self.league == WNBA_STRING else f'20{digit4}{digit5}-{digit4}{digit5 + 1}'

    @property
    def season_type(self):
        """
        3rd character in game id represent season type
        2 for reg season, 4 for playoffs
        """
        if self.game_id[2] == "4":
            return PLAYOFFS_STRING
        elif self.game_id[2] == "2":
            return REGULAR_SEASON_STRING



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='pbpstats',
    version='1.0.6',
    author='Darryl Blackport',
    author_email='darryl.blackport@gmail.com',
    description='A package to scrape and parse NBA, WNBA and G-League play-by-play data',
    license='MIT License',
    keywords=['basketball', 'NBA', 'WNBA', 'G-League', 'play-by-play', 'pbp'],
    url='https://github.com/dblackrun/pbpstats',
    packages=find_packages(),
    install_requires=['requests'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import json
import os

from pbpstats import (
    NBA_STRING, G_LEAGUE_STRING, D_LEAGUE_STRING, WNBA_STRING, PLAYOFFS_STRING,
    REGULAR_SEASON_STRING, NBA_GAME_ID_PREFIX, G_LEAGUE_GAME_ID_PREFIX, WNBA_GAME_ID_PREFIX
)
from pbpstats.data_loader.data_nba.file_loader import DataNbaFileLoader
from pbpstats.data_loader.data_nba.web_loader import DataNbaWebLoader
from pbpstats.resources.games.data_nba_game_item import DataNbaGameItem


class DataNbaScheduleLoader(DataNbaFileLoader, DataNbaWebLoader):
    data_provider = 'data_nba'
    resource = 'Games'
    parent_object = 'Season'

    def __init__(self, league, season, season_type, source, file_directory=None):
        self.league_string = league
        self.season_year = season.split('-')[0]
        self.file_directory = file_directory
        self.source = source
        self._load_data()
        self._make_game_items(season_type)

    def _load_data(self):
        source_method = getattr(self, f'_from_{self.source}')
        source_method()

    def _from_file(self):
        self.file_path = f'{self.file_directory}/schedule/data_{self.league_string}_{self.season_year}.json'
        self._load_data_from_file()

    def _from_web(self):
        league_url_part = NBA_STRING if self.league_string == G_LEAGUE_STRING else self.league_string
        league_part = D_LEAGUE_STRING if self.league_string == G_LEAGUE_STRING else self.league_string
        self.url = f'https://data.{league_url_part}.com/data/10s/v2015/json/mobile_teams/{league_part}/{self.season_year}/league/{self.league_id}_full_schedule.json'
        self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f'{self.file_directory}/schedule/data_{self.league_string}_{self.season_year}.json'
            with open(file_path, 'w') as outfile:
                json.dump(self.source_data, outfile)

    def _make_game_items(self, season_type):
        self.season_type = season_type
        self.items = [DataNbaGameItem(game) for item in self.data for game in item['mscd']['g'] if self._is_season_type(game)]

    def _is_season_type(self, game):
        if game['gid'][2] == '4' and self.season_type == PLAYOFFS_STRING:
            return True
        elif game['gid'][2] == '2' and self.season_type == REGULAR_SEASON_STRING:
            return True
        return False

    @property
    def data(self):
        return self.source_data['lscd']

    @property
    def league_id(self):
        if self.league_string == NBA_STRING:
            return NBA_GAME_ID_PREFIX
        elif self.league_string == WNBA_STRING:
            return WNBA_GAME_ID_PREFIX
        elif self.league_string == G_LEAGUE_STRING:
            return G_LEAGUE_GAME_ID_PREFIX



import inspect

import pbpstats.client as client


class Season(object):
    def __init__(self, league, season, season_type):
        self.league = league
        self.season = season
        self.season_type = season_type
        attributes = inspect.getmembers(self, lambda a: not(inspect.isroutine(a)))
        data_loaders = [a for a in attributes if a[0].endswith(client.DATA_LOADER_SUFFIX)]
        data_source_map = {a[0].replace(client.DATA_SOURCE_SUFFIX, ''): a[1] for a in attributes if a[0].endswith(client.DATA_SOURCE_SUFFIX)}
        for data_loader in data_loaders:
            attr_name = data_loader[0].replace(client.DATA_LOADER_SUFFIX, '')
            source = data_source_map[attr_name]
            data = data_loader[1](league, season, season_type, source, self.data_directory)
            resource_cls = getattr(self, attr_name)
            setattr(self, client.PATTERN.sub('_', attr_name).lower(), resource_cls(data.items))



from pbpstats.data_loader.data_nba.boxscore_loader import DataNbaBoxscoreLoader
from pbpstats.data_loader.data_nba.enhanced_pbp_loader import DataNbaEnhancedPbpLoader
from pbpstats.data_loader.data_nba.pbp_loader import DataNbaPbpLoader
from pbpstats.data_loader.data_nba.possessions_loader import DataNbaPossessionLoader
from pbpstats.data_loader.data_nba.schedule_loader import DataNbaScheduleLoader
from pbpstats.data_loader.stats_nba.boxscore_loader import StatsNbaBoxscoreLoader
from pbpstats.data_loader.stats_nba.enhanced_pbp_loader import StatsNbaEnhancedPbpLoader
from pbpstats.data_loader.stats_nba.game_finder_loader import StatsNbaGameFinderLoader
from pbpstats.data_loader.stats_nba.pbp_loader import StatsNbaPbpLoader
from pbpstats.data_loader.stats_nba.possessions_loader import StatsNbaPossessionLoader
from pbpstats.data_loader.stats_nba.scoreboard_loader import StatsNbaScoreboardLoader
from pbpstats.data_loader.stats_nba.shots_loader import StatsNbaShotsLoader

__all__ = [
    'DataNbaBoxscoreLoader',
    'DataNbaEnhancedPbpLoader',
    'DataNbaPbpLoader',
    'DataNbaPossessionLoader',
    'DataNbaScheduleLoader',
    'StatsNbaBoxscoreLoader',
    'StatsNbaEnhancedPbpLoader',
    'StatsNbaGameFinderLoader',
    'StatsNbaPbpLoader',
    'StatsNbaPossessionLoader',
    'StatsNbaScoreboardLoader',
    'StatsNbaShotsLoader',
]



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import StatsEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.jump_ball import JumpBall
from pbpstats.resources.enhanced_pbp.turnover import Turnover
from pbpstats.resources.enhanced_pbp.foul import Foul


class StatsJumpBall(JumpBall, StatsEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    def get_offense_team_id(self):
        """
        overrides method inherited from StatsEnhancedPbpItem
        """
        if self.next_event.clock == self.clock and isinstance(self.next_event, Turnover):
            return self.next_event.team_id
        if isinstance(self.next_event, Foul) and self.next_event.clock == self.clock:
            next_event = self.next_event.next_event
            if isinstance(next_event, Turnover) and not next_event.is_no_turnover and next_event.clock == self.clock:
                return next_event.team_id
        if self.count_as_possession:
            team_ids = list(self.current_players.keys())
            return team_ids[0] if team_ids[1] == self.team_id else team_ids[1]
        return self.team_id



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import DataEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.field_goal import FieldGoal


class DataFieldGoal(FieldGoal, DataEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    @property
    def shot_value(self):
        return 3 if '3pt Shot' in self.description else 2



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import StatsEnhancedPbpItem
from pbpstats.resources.enhanced_pbp.field_goal import FieldGoal


class StatsFieldGoal(FieldGoal, StatsEnhancedPbpItem):
    def __init__(self, *args):
        super().__init__(*args)

    @property
    def shot_value(self):
        return 3 if ' 3PT ' in self.description else 2

    def get_offense_team_id(self):
        """
        overrides method inherited from StatsEnhancedPbpItem
        """
        return self.team_id



VIDEO_EVENT_ASSET_BASE_URL = 'https://stats.nba.com/stats/videoeventsasset'
TODAYS_NBA_SCORES_URL = 'http://data.nba.com/data/5s/v2015/json/mobile_teams/nba/2019/scores/00_todays_scores.json'
TODAYS_WNBA_SCORES_URL = 'http://data.wnba.com/data/5s/v2015/json/mobile_teams/wnba/2019/scores/10_todays_scores.json'
TODAYS_G_LEAGUE_SCORES_URL = 'https://data.nba.com/data/10s/v2015/json/mobile_teams/dleague/2019/scores/20_todays_scores.json'
REQUEST_TIMEOUT = 10
USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.75.14 (KHTML, like Gecko) Version/7.0.3 Safari/7046A194A'
REFERER = "http://stats.nba.com/"
HEADERS = {
    'User-Agent': USER_AGENT,
    'Referer': REFERER,
    'Accept': 'application/json, text/plain, */*',
    'Accept-Language': 'en-US,en;q=0.5',
    'x-nba-stats-origin': 'stats',
    'x-nba-stats-token': 'true',
    'DNT': '1',
    'Connection': 'keep-alive',
    'Pragma': 'no-cache',
    'Cache-Control': 'no-cache'
}

NBA_STRING = 'nba'
G_LEAGUE_STRING = 'gleague'
WNBA_STRING = 'wnba'
D_LEAGUE_STRING = 'dleague'

NBA_GAME_ID_PREFIX = '00'
G_LEAGUE_GAME_ID_PREFIX = '20'
WNBA_GAME_ID_PREFIX = '10'

REGULAR_SEASON_STRING = 'Regular Season'
PLAYOFFS_STRING = 'Playoffs'

OFFENSIVE_ABBREVIATION_PREFIX = 'Off'
DEFENSIVE_ABBREVIATION_PREFIX = 'Def'
SECONDS_PLAYED_STRING = 'SecondsPlayed'
SECONDS_PLAYED_OFFENSE_STRING = f'{SECONDS_PLAYED_STRING}{OFFENSIVE_ABBREVIATION_PREFIX}'
SECONDS_PLAYED_DEFENSE_STRING = f'{SECONDS_PLAYED_STRING}{DEFENSIVE_ABBREVIATION_PREFIX}'
OFFENSIVE_POSSESSION_STRING = 'OffPoss'
DEFENSIVE_POSSESSION_STRING = 'DefPoss'
SECOND_CHANCE_STRING = 'SecondChance'
SECOND_CHANCE_SECONDS_PLAYED_OFFENSE_STRING = f'{SECOND_CHANCE_STRING}{SECONDS_PLAYED_OFFENSE_STRING}'
SECOND_CHANCE_SECONDS_PLAYED_DEFENSE_STRING = f'{SECOND_CHANCE_STRING}{SECONDS_PLAYED_DEFENSE_STRING}'

AT_RIM_CUTOFF = 4
SHORT_MID_RANGE_CUTOFF = 14
HEAVE_DISTANCE_CUTOFF = 40
HEAVE_TIME_CUTOFF = 2

TEAM_STAT_PLAYER_ID = '0'  # used as player id for team stats (ex team rebounds, shot clock violations)

AT_RIM_STRING = 'AtRim'
SHORT_MID_RANGE_STRING = 'ShortMidRange'
LONG_MID_RANGE_STRING = 'LongMidRange'
CORNER_3_STRING = 'Corner3'
ARC_3_STRING = 'Arc3'
UNKNOWN_SHOT_DISTANCE_STRING = 'UnknownDistance2pt'
FREE_THROW_STRING = 'FT'
FTS_MADE_STRING = 'FtsMade'
FT_1_PT_MADE_STRING = '1ptFtsMade'
FT_2_PT_MADE_STRING = '2ptFtsMade'
FT_3_PT_MADE_STRING = '3ptFtsMade'
TECHNICAL_FTS_MADE_STRING = 'TechFtsMade'
FTS_MISSED_STRING = 'FtsMissed'
FT_1_PT_MISSED_STRING = '1ptFtsMissed'
FT_2_PT_MISSED_STRING = '2ptFtsMissed'
FT_3_PT_MISSED_STRING = '3ptFtsMissed'
ASSISTS_STRING = 'Assists'
REBOUNDS_STRING = 'Rebounds'
ASSISTED_STRING = 'Assisted'
UNASSISTED_STRING = 'Unassisted'
MISSED_STRING = 'Missed'
BLOCKED_STRING = 'Blocked'
REBOUND_OPPORTUNITIES_STRING = 'ReboundOpportunities'
REBOUNDED_STRING = 'Rebounded'
REBOUNDED_OPPORTUNITIES_STRING = 'ReboundedOpportunities'  # this is for getting oreb% on a player's missed shots
ON_FLOOR_OFFENSIVE_REBOUND_STRING = 'OnFloorOffReb'  # used for calculating player usage
BLOCK_STRING = 'Block'
SELF_REBOUND_STRING = 'SelfOReb'

PUTBACKS_STRING = 'Putbacks'
HEAVE_MISSES_STRING = 'HeaveMisses'
HEAVE_MAKES_STRING = 'HeaveMakes'

# attributes for shot data
SECONDS_SINCE_OREB_STRING = 'SecondsSinceOReb'
OREB_SHOT_PLAYER_ID_STRING = 'OrebShotPlayerId'
OREB_REBOUND_PLAYER_ID_STRING = 'OrebReboundPlayerId'
START_SCORE_DIFFERENTIAL_STRING = 'StartScoreDifferential'
SHOT_DISTANCE_STRING = 'ShotDistance'
SECONDS_SINCE_PLAY_STARTED = 'SecondsSincePlayStarted'
IS_REGULAR_SEASON_STRING = 'IsRegularSeason'
SHOT_ANGLE_STRING = 'ShotAngle'
PUTBACK_STRING = 'Putback'

MAKE_STRING = 'Make'
MISS_STRING = 'Miss'
MADE_STRING = 'Made'

SHOT_VALUE_STRING = 'ShotValue'
OPPONENT_POINTS = 'OpponentPoints'

PERSONAL_FOUL_TYPE_STRING = 'Personal Fouls'
SHOOTING_FOUL_TYPE_STRING = 'Shooting Fouls'
LOOSE_BALL_FOUL_TYPE_STRING = 'Loose Ball Fouls'
OFFENSIVE_FOUL_TYPE_STRING = 'Offensive Fouls'
AWAY_FROM_PLAY_FOUL_TYPE_STRING = 'Away From Play Fouls'
CHARGE_FOUL_TYPE_STRING = 'Charge Fouls'
PERSONAL_BLOCK_TYPE_STRING = 'Personal Block Fouls'
PERSONAL_TAKE_TYPE_STRING = 'Personal Take Fouls'
SHOOTING_BLOCK_TYPE_STRING = 'Shooting Block Fouls'
CLEAR_PATH_FOUL_TYPE_STRING = 'Clear Path Fouls'
DEFENSIVE_3_SECONDS_FOUL_TYPE_STRING = 'Defensive 3 Seconds Violations'
FLAGRANT_1_FOUL_TYPE_STRING = 'Flagrant 1 Fouls'
FLAGRANT_2_FOUL_TYPE_STRING = 'Flagrant 2 Fouls'
DOUBLE_FOUL_TYPE_STRING = 'Double Fouls'
INBOUND_FOUL_TYPE_STRING = 'Inbound Fouls'

FOULS_DRAWN_TYPE_STRING = ' Drawn'

OFF_DEADBALL_STRING = 'OffDeadball'
OFF_LIVE_BALL_TURNOVER_STRING = 'OffLiveBallTurnover'
OFF_TIMEOUT_STRING = 'OffTimeout'
OFF_MADE_FT_STRING = f'Off{FREE_THROW_STRING}{MAKE_STRING}'
OFF_MISSED_FT_STRING = f'Off{FREE_THROW_STRING}{MISS_STRING}'

PENALTY_STRING = 'Penalty'
FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING = 'FinalMinutePenaltyTakeFoul'
PENALTY_POSSESSIONS_EXCLUDING_TAKE_FOUL_FTS = 'PenaltyOffPossExcludingTakeFouls'

PENALTY_FREE_THROW_STRING = 'Penalty Free Throw Trips'
THREE_POINT_AND1_FREE_THROW_STRING = '3pt And 1 Free Throw Trips'
TWO_POINT_AND1_FREE_THROW_STRING = '2pt And 1 Free Throw Trips'
THREE_POINT_SHOOTING_FOUL_FREE_THROW_STRING = '3pt Shooting Foul Free Throw Trips'
TWO_POINT_SHOOTING_FOUL_FREE_THROW_STRING = '2pt Shooting Foul Free Throw Trips'
TECHNICAL_FREE_THROW_STRING = 'Technical Free Throw Trips'

PLUS_MINUS_STRING = 'PlusMinus'
TIME_STRING = 'Time'

PLAYER_ID_STRING = 'PlayerId'
TEAM_ID_STRING = 'TeamId'
OPPONENT_TEAM_ID_STRING = 'OpponentTeamId'
LINEUP_ID_STRING = 'LineupId'
OPPONENT_LINEUP_ID_STRING = 'OpponentLineupId'

TOTAL_2PT_SHOT_DISTANCE_STRING = 'Total2ptShotDistance'
TOTAL_3PT_SHOT_DISTANCE_STRING = 'Total3ptShotDistance'
TOTAL_2PT_SHOTS_WITH_DISTANCE = 'Total2ptShotsWithDistance'
TOTAL_3PT_SHOTS_WITH_DISTANCE = 'Total3ptShotsWithDistance'

LIVEBALL_TURNOVERS_STRING = 'LiveBallTurnovers'
DEADBALL_TURNOVERS_STRING = 'DeadBallTurnovers'
STEALS_STRING = 'Steals'
TRAVELS_STRING = 'Travels'
THREE_SECOND_VIOLATION_TURNOVER_STRING = '3SecondViolations'
SHOT_CLOCK_VIOLATION_TURNOVER_STRING = 'ShotClockViolations'
STEP_OUT_OF_BOUNDS_TURNOVER_STRING = 'StepOutOfBoundsTurnovers'
LOST_BALL_TURNOVER_STRING = 'LostBallTurnovers'
LOST_BALL_OUT_OF_BOUNDS_TURNOVER_STRING = 'LostBallOutOfBoundsTurnovers'
BAD_PASS_TURNOVER_STRING = 'BadPassTurnovers'
BAD_PASS_OUT_OF_BOUNDS_TURNOVER_STRING = 'BadPassOutOfBoundsTurnovers'
LOST_BALL_STEAL_STRING = 'LostBallSteals'
BAD_PASS_STEAL_STRING = 'BadPassSteals'


OFFENSIVE_GOALTENDING_STRING = 'OffensiveGoaltends'
DEFENSIVE_GOALTENDING_STRING = 'DefensiveGoaltends'


KEYS_OFF_BY_FACTOR_OF_5_WHEN_AGGREGATING_FOR_TEAM_AND_LINEUPS = [
    OFFENSIVE_POSSESSION_STRING,
    DEFENSIVE_POSSESSION_STRING,
    SECONDS_PLAYED_DEFENSE_STRING,
    SECONDS_PLAYED_OFFENSE_STRING,
    PLUS_MINUS_STRING,
    OPPONENT_POINTS,
    FREE_THROW_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    AT_RIM_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    SHORT_MID_RANGE_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    LONG_MID_RANGE_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    ARC_3_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    CORNER_3_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    AT_RIM_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    SHORT_MID_RANGE_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    LONG_MID_RANGE_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    ARC_3_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    CORNER_3_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    FREE_THROW_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    AT_RIM_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    SHORT_MID_RANGE_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    LONG_MID_RANGE_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    ARC_3_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    CORNER_3_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    AT_RIM_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    SHORT_MID_RANGE_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    LONG_MID_RANGE_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    ARC_3_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    CORNER_3_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    SECOND_CHANCE_STRING + OFFENSIVE_POSSESSION_STRING,
    SECOND_CHANCE_STRING + DEFENSIVE_POSSESSION_STRING,
    PENALTY_STRING + OFFENSIVE_POSSESSION_STRING,
    PENALTY_STRING + DEFENSIVE_POSSESSION_STRING,
    PENALTY_STRING + SECONDS_PLAYED_OFFENSE_STRING,
    PENALTY_STRING + SECONDS_PLAYED_DEFENSE_STRING,
    FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING + OFFENSIVE_POSSESSION_STRING,
    FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING + DEFENSIVE_POSSESSION_STRING,
    PENALTY_POSSESSIONS_EXCLUDING_TAKE_FOUL_FTS,
    PENALTY_STRING + FREE_THROW_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + AT_RIM_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + SHORT_MID_RANGE_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + LONG_MID_RANGE_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + ARC_3_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + CORNER_3_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + AT_RIM_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + SHORT_MID_RANGE_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + LONG_MID_RANGE_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + ARC_3_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + CORNER_3_STRING + BLOCKED_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + FREE_THROW_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + AT_RIM_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + SHORT_MID_RANGE_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + LONG_MID_RANGE_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + ARC_3_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + CORNER_3_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + AT_RIM_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + SHORT_MID_RANGE_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + LONG_MID_RANGE_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + ARC_3_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING + CORNER_3_STRING + BLOCKED_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
]



import pbpstats.resources.enhanced_pbp.data_nba as event_types
from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import DataEnhancedPbpItem


class DataNbaEnhancedPbpFactory(object):
    def __init__(self):
        self.event_classes = {}
        self._load_event_classes()

    def _load_event_classes(self):
        event_classes = dict([(name, cls) for name, cls in event_types.__dict__.items() if isinstance(cls, type)])
        for _, event_cls in event_classes.items():
            if isinstance(event_cls.event_type, list):
                for event_type in event_cls.event_type:
                    self.event_classes[event_type] = event_cls
            else:
                self.event_classes[event_cls.event_type] = event_cls

    def get_event_class(self, event_type):
        return self.event_classes.get(event_type, DataEnhancedPbpItem)



import json
import os

from pbpstats import NBA_STRING, G_LEAGUE_STRING
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader
from pbpstats.resources.boxscore.stats_nba_boxscore_item import StatsNbaBoxscoreItem


class StatsNbaBoxscoreLoader(StatsNbaFileLoader, StatsNbaWebLoader):
    """
    class for loading boxscore items from stats.nba.com
    supports loading from api request or from file saved
    on disk with api response data
    """
    data_provider = 'stats_nba'
    resource = 'Boxscore'
    parent_object = 'Game'

    def __init__(self, game_id, source, file_directory):
        """
        game_id - string, stats.nba.com game id
        source - string, file or web
        file_directory - string, directory containing data saved on disk
            - required if loading from file
            - if not None when loading from web, response will be saved there
        """
        self.game_id = game_id
        self.file_directory = file_directory
        self.source = source
        self._load_data()
        self._make_boxscore_items()

    def _load_data(self):
        source_method = getattr(self, f'_from_{self.source}')
        source_method()

    def _from_file(self):
        self.file_path = f'{self.file_directory}/game_details/stats_boxscore_{self.game_id}.json'
        self._load_data_from_file()

    def _from_web(self):
        league_url_part = f'{G_LEAGUE_STRING}.{NBA_STRING}' if self.league == G_LEAGUE_STRING else self.league
        self.base_url = f'https://stats.{league_url_part}.com/stats/boxscoretraditionalv2'
        self.parameters = {
            'GameId': self.game_id,
            'StartPeriod': 0,
            'EndPeriod': 10,
            'RangeType': 2,
            'StartRange': 0,
            'EndRange': 55800
        }
        self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f'{self.file_directory}/game_details/stats_boxscore_{self.game_id}.json'
            with open(file_path, 'w') as outfile:
                json.dump(self.source_data, outfile)

    def _make_boxscore_items(self):
        self.items = [StatsNbaBoxscoreItem(item) for item in self.data]
        self.items += [StatsNbaBoxscoreItem(item) for item in self.make_list_of_dicts(results_set_index=1)]
